// key_openssl.c
//

#include <pktlab.h>

#include <assert.h>
#include <limits.h>
#include <stdint.h>
#include <string.h>

#include "private.h"

#include "key_openssl.h"
#include "util_openssl.h"

#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <openssl/sha.h>
#include <openssl/x509.h>

//
// INTERNAL MACROS
//

//
// INTERNAL FUNCTION DECLARATIONS
//

static BIO * create_BIO(const void * ptr, uint_fast32_t len);
static char * create_ppcopy(const void * pp);

//
// EXPORTED FUNCTION DEFINITIONS
//

// pktlab.h functions
struct pktlab_publickey * pktlab_create_publickey(void) {
    struct pktlab_publickey * k;

    trace("%s()", __func__);

    k = safe_malloc(sizeof(struct pktlab_publickey));
    k->key = NULL;
    return k;
}

int pktlab_load_publickey_pem (
    const char * pem, uint_fast32_t pemlen,
    struct pktlab_publickey * k)
{
    int ret;
    BIO * bio;

    trace("%s(pemlen:%u)", __func__, (unsigned) pemlen);

    if (pem == NULL || pemlen == 0 || k == NULL || pemlen > INT_MAX) {
        debug("%s(pem:%p,pemlen:%u,k:%p): bad argument",
            __func__, pem, (unsigned) pemlen, k);
        return -1;
    }

    ret = 0;
    bio = create_BIO(pem, pemlen);
    ERR_clear_error();
    k->key = PEM_read_bio_PUBKEY(bio, NULL, NULL, NULL);
    if (k->key == NULL) {
        error_debug_msg("PEM_read_bio_PUBKEY");
        ret = -1;
    }

    BIO_free_all(bio);
    return ret;
}

int pktlab_load_publickey_der (
    const uint8_t * der, uint_fast32_t derlen,
    struct pktlab_publickey * k)
{
    int ret;
    const uint8_t * ptr;

    trace("%s(derlen:%u)", __func__, (unsigned) derlen);

    if (der == NULL || derlen == 0 || k == NULL || derlen > LONG_MAX) {
        debug("%s(der:%p,derlen:%u,k:%p): bad argument",
            __func__, der, (unsigned) derlen, k);
        return -1;
    }

    ret = 0;
    ptr = der;
    ERR_clear_error();
    k->key = d2i_PUBKEY(NULL, &ptr, derlen);
    if (k->key == NULL) {
        error_debug_msg("d2i_PUBKEY");
        ret = -1;
    }

    return ret;
}

int pktlab_load_publickey_raw (
    const uint8_t * bytes, uint_fast32_t len,
    struct pktlab_publickey * k)
{
    int ret;

    trace("%s(len:%u)", __func__, (unsigned) len);

    if (bytes == NULL || len == 0 || k == NULL || len > SIZE_MAX) {
        debug("%s(bytes:%p,len:%u,k:%p): bad argument",
            __func__, bytes, (unsigned) len, k);
        return -1;
    }

    ret = 0;
    ERR_clear_error();
    k->key = EVP_PKEY_new_raw_public_key(EVP_PKEY_ED25519, NULL, bytes, len);
    if (k->key == NULL) {
        error_debug_msg("EVP_PKEY_new_raw_public_key");
        ret = -1;
    }

    return ret;
}

int pktlab_cleanup_publickey(struct pktlab_publickey * k)
{
    trace("%s(...)", __func__);

    if (k == NULL) {
        debug("%s(k:%p): bad argument", __func__, k);
        return -1;
    }

    EVP_PKEY_free(k->key);
    free(k);
    return 0;
}

int pktlab_get_publickey_bytes (
    const struct pktlab_publickey * k,
    void * buf, uint_fast32_t buflen)
{
    size_t len;

    trace("%s(buflen:%u)", __func__, (unsigned) buflen);

    if (k == NULL || k->key == NULL ||
        buf == NULL || buflen < PKTLAB_PUBLICKEY_LEN) {
        debug("%s(k:%p,buf:%p,buflen:%u): bad argument",
            __func__, k, buf, (unsigned) buflen);
        return -1;
    }

    len = (buflen > SIZE_MAX) ? SIZE_MAX : buflen;
    ERR_clear_error();
    if (EVP_PKEY_get_raw_public_key(k->key, buf, &len) != 1) {
        error_debug_msg("EVP_PKEY_get_raw_public_key");
        return -1;
    }

    assert(len == PKTLAB_PUBLICKEY_LEN);

    return 0;
}

struct pktlab_privatekey * pktlab_create_privatekey(void) {
    struct pktlab_privatekey * k;

    trace("%s()", __func__);

    k = safe_malloc(sizeof(struct pktlab_privatekey));
    k->key = NULL;
    return k;
}

int pktlab_load_privatekey_pem (
    const char * pem, uint_fast32_t pemlen,
    const char * passphrase, struct pktlab_privatekey * k)
{
    int ret;
    BIO * bio;
    char * ppcopy;

    trace("%s(pemlen:%u,passphrase:%s)",
        __func__, (unsigned) pemlen,
        (passphrase == NULL) ? "NULL":"REDACTED");

    if (pem == NULL || pemlen == 0 || k == NULL || pemlen > INT_MAX) {
        debug("%s(pem:%p,pemlen:%u,passphrase:%p,k:%p): bad argument",
            __func__, pem, (unsigned) pemlen, passphrase, k);
        return -1;
    }

    ret = 0;
    bio = create_BIO(pem, pemlen);
    ppcopy = create_ppcopy(passphrase);
    ERR_clear_error();
    k->key = PEM_read_bio_PrivateKey(bio, NULL, NULL, ppcopy);

    if (k->key == NULL) {
        error_debug_msg("PEM_read_bio_PrivateKey");
        ret = -1;
    }

    BIO_free_all(bio);
    free(ppcopy);
    return ret;
}

int pktlab_load_privatekey_der (
    const uint8_t * der, uint_fast32_t derlen,
    const char * passphrase, struct pktlab_privatekey * k)
{
    int ret;
    BIO * bio;
    char * ppcopy;
    PKCS8_PRIV_KEY_INFO * kinfo;

    trace("%s(derlen:%u,passphrase:%s)",
        __func__, (unsigned) derlen,
        (passphrase == NULL) ? "NULL":"REDACTED");

    if (der == NULL || derlen == 0 || k == NULL || derlen > INT_MAX) {
        debug("%s(der:%p,derlen:%u,passphrase:%p,k:%p): bad argument",
            __func__, der, (unsigned) derlen, passphrase, k);
        return -1;
    }

    ret = 0;
    bio = create_BIO(der, derlen);

    if (passphrase != NULL) {
        ppcopy = create_ppcopy(passphrase);
        ERR_clear_error();
        k->key = d2i_PKCS8PrivateKey_bio(bio, NULL, NULL, ppcopy);
        if (k->key == NULL) {
            error_debug_msg("d2i_PKCS8PrivateKey_bio");
            ret = -1;
        }

        free(ppcopy);
    } else {
        ERR_clear_error();
        kinfo = d2i_PKCS8_PRIV_KEY_INFO_bio(bio, NULL);
        if (kinfo == NULL) {
            error_debug_msg("d2i_PKCS8_PRIV_KEY_INFO_bio");
            ret = -1;
        } else {
            ERR_clear_error();
            k->key = EVP_PKCS82PKEY(kinfo);
            if (k->key == NULL) {
                error_debug_msg("EVP_PKCS82PKEY");
                ret = -1;
            }

            PKCS8_PRIV_KEY_INFO_free(kinfo);
        }
    }

    BIO_free_all(bio);
    return ret;
}

int pktlab_load_privatekey_raw (
    const uint8_t * bytes, uint_fast32_t len,
    struct pktlab_privatekey * k)
{
    int ret;

    trace("%s(len:%u)", __func__, (unsigned) len);

    if (bytes == NULL || len == 0 || k == NULL || len > SIZE_MAX) {
        debug("%s(bytes:%p,len:%u,k:%p): bad argument",
            __func__, bytes, (unsigned) len, k);
        return -1;
    }

    ret = 0;
    ERR_clear_error();
    k->key = EVP_PKEY_new_raw_private_key(EVP_PKEY_ED25519, NULL, bytes, len);
    if (k->key == NULL) {
        error_debug_msg("EVP_PKEY_new_raw_private_key");
        ret = -1;
    }

    return ret;
}

int pktlab_cleanup_privatekey(struct pktlab_privatekey * k)
{
    trace("%s(...)", __func__);

    if (k == NULL) {
        debug("%s(k:%p): bad argument", __func__, k);
        return -1;
    }

    EVP_PKEY_free(k->key);
    free(k);
    return 0;
}

int pktlab_get_privatekey_bytes (
    const struct pktlab_privatekey * k,
    void * buf, uint_fast32_t buflen)
{
    size_t len;

    trace("%s(buflen:%u)", __func__, (unsigned) buflen);

    if (k == NULL || k->key == NULL ||
        buf == NULL || buflen < PKTLAB_PRIVATEKEY_LEN) {
        debug("%s(k:%p,buf:%p,buflen:%u): bad argument",
            __func__, k, buf, (unsigned) buflen);
        return -1;
    }

    len = (buflen > SIZE_MAX) ? SIZE_MAX : buflen;
    ERR_clear_error();
    if (EVP_PKEY_get_raw_private_key(k->key, buf, &len) != 1) {
        error_debug_msg("EVP_PKEY_get_raw_public_key");
        return -1;
    }

    assert(len == PKTLAB_PRIVATEKEY_LEN);

    return 0;
}

int pktlab_get_key_identifier (
    const struct pktlab_publickey * k,
    void * buf, uint_fast32_t buflen)
{
    uint8_t rawkey[PKTLAB_PUBLICKEY_LEN];

    trace("%s(buflen:%u)", __func__, (unsigned) buflen);

    assert(PKTLAB_KEYID_LEN == SHA256_DIGEST_LENGTH);

    if (k == NULL || buf == NULL || buflen < PKTLAB_KEYID_LEN) {
        debug("%s(k:%p,buf:%p,buflen:%u): bad argument",
            __func__, k, buf, (unsigned) buflen);
        return -1;
    }

    if (pktlab_get_publickey_bytes(
            k, rawkey, sizeof(rawkey)) < 0)
        return -1;

    ERR_clear_error();
    if (SHA256(rawkey, sizeof(rawkey), buf) == NULL) {
        error_debug_msg("SHA256");
        return -1;
    }

    return 0;
}

struct pktlab_publickey * pktlab_get_publickey_from_privatekey (
    const struct pktlab_privatekey * k)
{
    BIO * bio;
    struct pktlab_publickey * ret;

    trace("%s(...)", __func__);

    if (k == NULL || k->key == NULL) {
        debug("%s(k:%p): bad argument", __func__, k);
        return NULL;
    }

    ret = pktlab_create_publickey();

    bio = BIO_new(BIO_s_mem());
    if (PEM_write_bio_PUBKEY(bio, k->key) != 1)
        goto fail;

    ret->key = PEM_read_bio_PUBKEY(bio, NULL, NULL, NULL);
    if (ret->key == NULL)
        goto fail;

    BIO_free_all(bio);
    return ret;

fail:
    BIO_free_all(bio);
    pktlab_cleanup_publickey(ret);
    return NULL;
}

// key_openssl.h functions
// Weird work around for copying Ed25519 EVP_PKEY with d2i and i2d method
int dup_key (
    const struct pktlab_publickey * k,
    struct pktlab_publickey ** kdup)
{
    int len;
    unsigned char * pp = NULL;
    unsigned char * pdup;

    assert(k != NULL && kdup != NULL);

    *kdup = pktlab_create_publickey();

    ERR_clear_error();
    len = i2d_PUBKEY(k->key, &pp);
    if (len < 0) {
        error_debug_msg("i2d_PUBKEY");
        goto fail;
    }

    pdup = pp;
    ERR_clear_error();
    (*kdup)->key = d2i_PUBKEY(
        NULL, (const unsigned char **) &pdup, len);
    if ((*kdup)->key == NULL) {
        error_debug_msg("d2i_PUBKEY");
        goto fail;
    }

    OPENSSL_free(pp);
    return 0;

fail:
    pktlab_cleanup_publickey(*kdup);
    if (pp != NULL)
        OPENSSL_free(pp);
    return -1;
}

//
// INTERNAL FUNCTION DEFINITION
//

BIO * create_BIO(const void * ptr, uint_fast32_t len) {
    BIO * ret;

    assert(ptr != NULL && len != 0 && len <= INT_MAX);

    ret = BIO_new_mem_buf(ptr, len);
    return ret;
}

char * create_ppcopy(const void * pp) {
    char * ppcopy;
    size_t pplen;

    if (pp != NULL) {
        pplen = strlen(pp); // passphrase must be null terminated
        ppcopy = safe_malloc(pplen+1);
        strncpy(ppcopy, pp, pplen+1);
    } else { // suppress openssl terminal prompt for passphrase if not provided
        ppcopy = safe_malloc(1);
        *ppcopy = '\0';
    }

    return ppcopy;
}