// linked_list.c
// Linked list C implementation.
//

#include "linked_list.h"
#include "private.h"

#include <assert.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

//
// INTERNAL TYPE DEFINITION
//

struct ll_node {
    void * data;
    struct ll_node * next;
};

struct linked_list {
    struct ll_node * head;
    struct ll_node * tail;
    uint_fast32_t len;
};

//
// EXPORTED FUNCTION DEFINITIONS
//

struct linked_list * create_ll(void) {
    struct linked_list * ll;

    ll = safe_malloc(sizeof(struct linked_list));
    memset(ll, 0, sizeof(struct linked_list));

    ll->head = ll->tail = NULL;
    ll->len = 0;

    return ll;
}

int push_ll(void * data, struct linked_list * ll) {
    struct ll_node * new;

    assert(ll != NULL);

    new = safe_malloc(sizeof(struct ll_node));
    new->data = data;
    new->next = NULL;

    if (ll->len == 0) {
        ll->head = ll->tail = new;
    } else {
        ll->tail->next = new;
        ll->tail = new;
    }

    ++ll->len;
    return 0;
}

void cleanup_ll (
    struct linked_list * ll, freedata_t free_func, void * aux)
{
    struct ll_node * node;
    struct ll_node * next;

    assert(ll != NULL);

    node = get_head(ll);
    while (node != NULL) {
        if (free_func != NULL)
            free_func(get_data(node), aux);

        next = get_next(node);
        free(node);
        node = next;
    }

    free(ll);
    return;
}

struct linked_list * copy_ll (
    const struct linked_list * ll)
{
    struct linked_list * ret;
    struct ll_node * node;
    void * data;

    assert(ll != NULL);

    ret = create_ll();
    if (ret == NULL)
        goto fail;

    for (node = get_head(ll); node != NULL; node = get_next(node)) {
        data = get_data(node);
        if (push_ll(data, ret) < 0)
            goto fail_ll;
    }

    return ret;

fail_ll:
    cleanup_ll(ret, NULL, NULL); // don't free the passed ll content
fail:
    return NULL;
}

uint_fast32_t len_ll(const struct linked_list * ll) {
    assert(ll != NULL);
    return ll->len;
}

struct ll_node * get_head(const struct linked_list * ll) {
    assert(ll != NULL);
    return ll->head;
}

struct ll_node * get_next(const struct ll_node * node) {
    assert(node != NULL);
    return node->next;
}

struct ll_node * get_node (
    const struct linked_list * ll, uint_fast32_t indx)
{
    uint_fast32_t i;
    struct ll_node * node;

    assert(ll != NULL);

    if (len_ll(ll) <= indx) // len_ll(ll) == 0 is automatically considered
        return NULL;

    i = 0;
    node = get_head(ll);
    while (true) {
        if (i == indx)
            return node;

        ++i;
        node = get_next(node);
    }
}

void * get_data(const struct ll_node * node) {
    assert(node != NULL);
    return node->data;
}

int remove_node(struct linked_list * ll, uint_fast32_t indx, void ** data) {
    uint_fast32_t i;
    struct ll_node * node;
    struct ll_node * pre;

    assert(ll != NULL);

    if (indx >= len_ll(ll))
        return -1;

    if (indx == 0) {
        if (data != NULL)
            *data = get_data(ll->head);

        node = ll->head;
        ll->head = get_next(node);
        if (len_ll(ll) == 1)
            ll->tail = NULL;

        free(node);
        --ll->len;
        return 0;
    }

    // len_ll(ll) >= 2 to reach here

    i = 1;
    pre = get_head(ll);
    node = get_next(pre);
    while (true) {
        if (i == indx) {
            if (data != NULL)
                *data = get_data(node);

            pre->next = get_next(node);
            if (len_ll(ll)-1 == indx) // removing tail
                ll->tail = pre;

            free(node);
            --ll->len;
            break;
        }

        ++i;
        pre = node;
        node = get_next(node);
    }

    return 0;
}

int replace_data(struct ll_node * node, void ** olddata, void * newdata) {
    assert(node != NULL);

    if (olddata != NULL)
        *olddata = get_data(node);

    node->data = newdata;
    return 0;
}