// pktlab/message.c
//

#include <pktctrl.h>
#include <pktlab.h>

#include <assert.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <inttypes.h>
#include <arpa/inet.h>
#include <sys/uio.h>

#include "channel.h"
#include "private.h"

//
// INTERNAL FUNCTION DECLARATIONS
//

// Message decoders are called by pktlab_decode_message to decode a raw
// sequence of bytes. On entry, msg is a pointer to an uninitialized
// pktlab_message structure and ptr is a pointer to len bytes of memory
// containing the received message, starting with the header. Headers are
// dispatched by pktlab_decode_message using the message type supplied in
// the header. A decoder returns true if decoding succeeded and false
// otherwise. The msg structure may be initialized with pointers into the
// len bytes of memory starting with ptr. Decoders must not return any
// pointers to dynamically allocated memory. That is, any pointers in msg
// must be NULL or must point into memory pointed to by ptr.

typedef bool (*message_decoder_t)(struct pktlab_message *restrict msg,
                                  const void *restrict ptr, uint_fast32_t len);

static void make_raw_message(struct pktlab_message *restrict msg,
                             const void *restrict rawptr, uint_fast32_t rawlen,
                             bool malformed);
static bool decode_result_message(struct pktlab_message *restrict msg,
                                  const void *restrict ptr, uint_fast32_t len);
static bool decode_cdata_message(struct pktlab_message *restrict msg,
                                 const void *restrict ptr, uint_fast32_t len);
static bool decode_start_message(struct pktlab_message *restrict msg,
                                 const void *restrict ptr, uint_fast32_t len);
static bool decode_end_message(struct pktlab_message *restrict msg,
                               const void *restrict ptr, uint_fast32_t len);
static bool decode_yield_message(struct pktlab_message *restrict msg,
                                 const void *restrict ptr, uint_fast32_t len);
static bool decode_cont_message(struct pktlab_message *restrict msg,
                                const void *restrict ptr, uint_fast32_t len);
static bool decode_mread_message(struct pktlab_message *restrict msg,
                                 const void *restrict ptr, uint_fast32_t len);
static bool decode_mwrite_message(struct pktlab_message *restrict msg,
                                  const void *restrict ptr, uint_fast32_t len);
static bool decode_nopen_message(struct pktlab_message *restrict msg,
                                 const void *restrict ptr, uint_fast32_t len);
static bool decode_nclose_message(struct pktlab_message *restrict msg,
                                  const void *restrict ptr, uint_fast32_t len);
static bool decode_nsend_message(struct pktlab_message *restrict msg,
                                 const void *restrict ptr, uint_fast32_t len);
static bool decode_ncap_message(struct pktlab_message *restrict msg,
                                const void *restrict ptr, uint_fast32_t len);
static bool decode_nctl_message(struct pktlab_message *restrict msg,
                                const void *restrict ptr, uint_fast32_t len);
static bool decode_mdata_message(struct pktlab_message *restrict msg,
                                 const void *restrict ptr, uint_fast32_t len);
static bool decode_ntag_message(struct pktlab_message *restrict msg,
                                const void *restrict ptr, uint_fast32_t len);
static bool decode_nstat_message(struct pktlab_message *restrict msg,
                                 const void *restrict ptr, uint_fast32_t len);
static bool decode_ndata_message(struct pktlab_message *restrict msg,
                                 const void *restrict ptr, uint_fast32_t len);
static bool decode_ndrop_message(struct pktlab_message *restrict msg,
                                 const void *restrict ptr, uint_fast32_t len);
static bool decode_xcert_message(struct pktlab_message *restrict msg,
                                 const void *restrict ptr, uint_fast32_t len);
static bool decode_xpub_message(struct pktlab_message *restrict msg,
                                const void *restrict ptr, uint_fast32_t len);
static bool decode_xsub_message(struct pktlab_message *restrict msg,
                                const void *restrict ptr, uint_fast32_t len);
static bool decode_suspd_message(struct pktlab_message *restrict msg,
                                 const void *restrict ptr, uint_fast32_t len);
static bool decode_resumd_message(struct pktlab_message *restrict msg,
                                  const void *restrict ptr, uint_fast32_t len);
static bool decode_xnotify_message(struct pktlab_message *restrict msg,
                                   const void *restrict ptr, uint_fast32_t len);
static bool decode_xreg_message(struct pktlab_message *restrict msg,
                                const void *restrict ptr, uint_fast32_t len);

// Message encoders are called by pktlab_encode_message to encode a message
// given as a pklab_message structure into up to PKTLAB_ENCODE_IOVCNT
// non-contiguous sequences of bytes. On entry, msg points to an
// initialized pklab_message structure and iov points to an array of
// PKTLAB_ENCODE_IOVCNT struct iovec elements. All but the first element of
// iov are initialized to 0. The first element is initialized so that
// iov[0].iov_base points to a buffer of size at least PKTLAB_ENCODE_BUFSZ
// and iov[0].iov_len is PKTLAB_HLEN. The encoder may use the remaining
// PKTLAB_ENCODE_BUFSZ - PKTLAB_HLEN bytes of the buffer to encode parts of
// the message as well as the remaining elements of iov. On return, iov
// should be filled in so that, excluding the header, the regions of memory
// defined by the iov array encode a valid message. The message header will
// be initialized by pktlab_encode_message after return from the encoder.
// Encoders must not return any pointers to dynamically allocated memory,
// however, iov_base pointers may point to memory regions referenced by msg.
// An encoder must return the number of iovec elements used (including the
// first) if encoding was successful and -1 otherwise. In the latter case,
// errno must be set to indicate the error.

typedef int (*message_encoder_t)(const struct pktlab_message *restrict msg,
                                 struct iovec *restrict iov);

static int encode_result_message(const struct pktlab_message *restrict msg,
                                 struct iovec *restrict iov);
static int encode_cdata_message(const struct pktlab_message *restrict msg,
                                struct iovec *restrict iov);
static int encode_start_message(const struct pktlab_message *restrict msg,
                                struct iovec *restrict iov);
static int encode_end_message(const struct pktlab_message *restrict msg,
                              struct iovec *restrict iov);
static int encode_yield_message(const struct pktlab_message *restrict msg,
                                struct iovec *restrict iov);
static int encode_cont_message(const struct pktlab_message *restrict msg,
                               struct iovec *restrict iov);
static int encode_mread_message(const struct pktlab_message *restrict msg,
                                struct iovec *restrict iov);
static int encode_mwrite_message(const struct pktlab_message *restrict msg,
                                 struct iovec *restrict iov);
static int encode_nopen_message(const struct pktlab_message *restrict msg,
                                struct iovec *restrict iov);
static int encode_nclose_message(const struct pktlab_message *restrict msg,
                                 struct iovec *restrict iov);
static int encode_nsend_message(const struct pktlab_message *restrict msg,
                                struct iovec *restrict iov);
static int encode_ncap_message(const struct pktlab_message *restrict msg,
                               struct iovec *restrict iov);
static int encode_nctl_message(const struct pktlab_message *restrict msg,
                               struct iovec *restrict iov);
static int encode_mdata_message(const struct pktlab_message *restrict msg,
                                struct iovec *restrict iov);
static int encode_ntag_message(const struct pktlab_message *restrict msg,
                               struct iovec *restrict iov);
static int encode_nstat_message(const struct pktlab_message *restrict msg,
                                struct iovec *restrict iov);
static int encode_ndata_message(const struct pktlab_message *restrict msg,
                                struct iovec *restrict iov);
static int encode_ndrop_message(const struct pktlab_message *restrict msg,
                                struct iovec *restrict iov);
static int encode_xcert_message(const struct pktlab_message *restrict msg,
                                struct iovec *restrict iov);
static int encode_xpub_message(const struct pktlab_message *restrict msg,
                               struct iovec *restrict iov);
static int encode_xsub_message(const struct pktlab_message *restrict msg,
                               struct iovec *restrict iov);
static int encode_suspd_message(const struct pktlab_message *restrict msg,
                                struct iovec *restrict iov);
static int encode_resumd_message(const struct pktlab_message *restrict msg,
                                 struct iovec *restrict iov);
static int encode_xnotify_message(const struct pktlab_message *restrict msg,
                                  struct iovec *restrict iov);
static int encode_xreg_message(const struct pktlab_message *restrict msg,
                               struct iovec *restrict iov);

static bool is_good_xdescr(const void *ptr, uint_fast32_t len);
static bool is_good_dattr(const void *ptr, uint_fast32_t len);
static bool is_good_chlist(const void *ptr, uint_fast32_t len);
static bool is_good_snikey(const void *ptr, uint_fast32_t len);

//
// INTERNAL MACRO DEFINITIONS
//

#define RETURN_IF(c, v)                                                        \
    if (c)                                                                     \
    return (v)

//
// EXPORTED FUNCTION DEFINITIONS
//

size_t pktlab_decode_message(struct pktlab_message *msg, const void *ptr,
                             size_t len) {
    static const message_decoder_t decoders[] = {
        [PKTLAB_RESULT_MESSAGE] = &decode_result_message,
        [PKTLAB_CDATA_MESSAGE] = &decode_cdata_message,
        [PKTLAB_START_MESSAGE] = &decode_start_message,
        [PKTLAB_END_MESSAGE] = &decode_end_message,
        [PKTLAB_YIELD_MESSAGE] = &decode_yield_message,
        [PKTLAB_CONT_MESSAGE] = &decode_cont_message,
        [PKTLAB_MREAD_MESSAGE] = &decode_mread_message,
        [PKTLAB_MWRITE_MESSAGE] = &decode_mwrite_message,
        [PKTLAB_NOPEN_MESSAGE] = &decode_nopen_message,
        [PKTLAB_NCLOSE_MESSAGE] = &decode_nclose_message,
        [PKTLAB_NSEND_MESSAGE] = &decode_nsend_message,
        [PKTLAB_NCAP_MESSAGE] = &decode_ncap_message,
        [PKTLAB_NCTL_MESSAGE] = &decode_nctl_message,
        [PKTLAB_MDATA_MESSAGE] = &decode_mdata_message,
        [PKTLAB_NTAG_MESSAGE] = &decode_ntag_message,
        [PKTLAB_NSTAT_MESSAGE] = &decode_nstat_message,
        [PKTLAB_NDATA_MESSAGE] = &decode_ndata_message,
        [PKTLAB_NDROP_MESSAGE] = &decode_ndrop_message,
        [PKTLAB_XCERT_MESSAGE] = &decode_xcert_message,
        [PKTLAB_XPUB_MESSAGE] = &decode_xpub_message,
        [PKTLAB_XSUB_MESSAGE] = &decode_xsub_message,
        [PKTLAB_SUSPD_MESSAGE] = &decode_suspd_message,
        [PKTLAB_RESUMD_MESSAGE] = &decode_resumd_message,
        [PKTLAB_XNOTIFY_MESSAGE] = &decode_xnotify_message,
        [PKTLAB_XREG_MESSAGE] = &decode_xreg_message};

    message_decoder_t decoder;
    bool malformed = false;
    uint_fast8_t msgtype;
    uint_fast32_t msglen;
    const void *msgptr;

    trace("pktlab_decode_message(len:%zu)", len);

    if (len < PKTLAB_HLEN)
        return 0;

    msgtype = *(const uint8_t *)ptr;
    msglen = get24n(ptr + 1);
    msgptr = ptr + PKTLAB_HLEN;

    debug("msgtype: %d", msgtype);
    debug("raw: %d %d %d %d", *((uint8_t *)ptr), *((uint8_t *)ptr + 1),
          *((uint8_t *)ptr + 2), *((uint8_t *)ptr + 3));
    if (msglen > len - PKTLAB_HLEN)
        return 0;

    memset(msg, 0, sizeof(struct pktlab_message));

    if (msgtype < sizeof(decoders) / sizeof(message_decoder_t)) {
        if ((decoder = decoders[msgtype])) {
            if (!decoder(msg, msgptr, msglen))
                malformed = true;
            else
                goto done;
        }
    }

    make_raw_message(msg, ptr, PKTLAB_HLEN + msglen, malformed);

done:
    return PKTLAB_HLEN + msglen;
}

int pktlab_encode_message(const struct pktlab_message *restrict msg,
                          void *restrict buf, struct iovec *restrict iov) {
    static const message_encoder_t encoders[] = {
        [PKTLAB_RESULT_MESSAGE] = &encode_result_message,
        [PKTLAB_CDATA_MESSAGE] = &encode_cdata_message,
        [PKTLAB_START_MESSAGE] = &encode_start_message,
        [PKTLAB_END_MESSAGE] = &encode_end_message,
        [PKTLAB_YIELD_MESSAGE] = &encode_yield_message,
        [PKTLAB_CONT_MESSAGE] = &encode_cont_message,
        [PKTLAB_MREAD_MESSAGE] = &encode_mread_message,
        [PKTLAB_MWRITE_MESSAGE] = &encode_mwrite_message,
        [PKTLAB_NOPEN_MESSAGE] = &encode_nopen_message,
        [PKTLAB_NCLOSE_MESSAGE] = &encode_nclose_message,
        [PKTLAB_NSEND_MESSAGE] = &encode_nsend_message,
        [PKTLAB_NCAP_MESSAGE] = &encode_ncap_message,
        [PKTLAB_NCTL_MESSAGE] = &encode_nctl_message,
        [PKTLAB_MDATA_MESSAGE] = &encode_mdata_message,
        [PKTLAB_NTAG_MESSAGE] = &encode_ntag_message,
        [PKTLAB_NSTAT_MESSAGE] = &encode_nstat_message,
        [PKTLAB_NDATA_MESSAGE] = &encode_ndata_message,
        [PKTLAB_NDROP_MESSAGE] = &encode_ndrop_message,
        [PKTLAB_XCERT_MESSAGE] = &encode_xcert_message,
        [PKTLAB_XPUB_MESSAGE] = &encode_xpub_message,
        [PKTLAB_XSUB_MESSAGE] = &encode_xsub_message,
        [PKTLAB_SUSPD_MESSAGE] = &encode_suspd_message,
        [PKTLAB_RESUMD_MESSAGE] = &encode_resumd_message,
        [PKTLAB_XNOTIFY_MESSAGE] = &encode_xnotify_message,
        [PKTLAB_XREG_MESSAGE] = &encode_xreg_message};

    message_encoder_t encoder;
    uint_fast32_t msglen;
    int iovcnt;
    int i;

    trace("pktlab_encode_message(msg:{type:%d})", (int)msg->type);

    memset(iov, 0, PKTLAB_ENCODE_IOVCNT * sizeof(struct iovec));

    if (msg->type < sizeof(encoders) / sizeof(message_encoder_t))
        encoder = encoders[msg->type];
    else
        encoder = NULL;

    if (encoder != NULL) {
        iov[0].iov_base = buf;
        iov[0].iov_len = PKTLAB_HLEN;

        iovcnt = encoder(msg, iov);
        RETURN_IF(iovcnt <= 0, iovcnt);

        msglen = -PKTLAB_HLEN;
        for (i = 0; i < iovcnt; i++)
            msglen += iov[i].iov_len;

        if (msglen > UINT32_C(0xffffff)) {
            errno = EMSGSIZE;
            return -1;
        }

        set32n(buf, ((uint_fast32_t)msg->type << 24) | msglen);
        debug("encoded header: %08" PRIxFAST32, get32n(buf));

        return iovcnt;
    }

    if (msg->type == PKTLAB_UNDEF_MESSAGE) {
        iov[0].iov_base = (void *)msg->raw.ptr;
        iov[0].iov_len = msg->raw.len;
        return 1;
    }

    errno = EINVAL;
    return -1;
}

extern struct pktlab_message *
pktlab_create_result_message(enum pktlab_status status) {
    struct pktlab_message *msg;
    msg = safe_malloc(sizeof(struct pktlab_message));
    memset(msg, 0, sizeof(struct pktlab_message));
    msg->type = PKTLAB_RESULT_MESSAGE;
    msg->result.errid = status;
    return msg;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

void make_raw_message(struct pktlab_message *msg, const void *rawptr,
                      uint_fast32_t rawlen, bool malformed) {
    trace("make_raw_message(rawlen:%" PRIuFAST32 ", malformed:%s)", rawlen,
          malformed ? "true" : "false");

    msg->type = PKTLAB_UNDEF_MESSAGE;
    msg->raw.malformed = malformed;
    msg->raw.ptr = rawptr;
    msg->raw.len = rawlen;
}

bool decode_result_message(struct pktlab_message *msg, const void *ptr,
                           uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    RETURN_IF(!((len == 0) || (len == 2)), false);
    msg->type = PKTLAB_RESULT_MESSAGE;

    if (len == 0) {
        msg->result.errid = 0;
    } else {
        msg->result.errid = get16n(ptr);
    }
    return true;
}

int encode_result_message(const struct pktlab_message *msg, struct iovec *iov) {
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;
    trace("%s(msg:{errid:%d})", __func__, (int)msg->result.errid);

    if (msg->result.errid != 0) {
        set16n(ptr, msg->result.errid);
        iov[0].iov_len += 2;
    }
    return 1;
}

bool decode_cdata_message(struct pktlab_message *msg, const void *ptr,
                          uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    RETURN_IF(len < 4, false);

    msg->type = PKTLAB_CDATA_MESSAGE;
    msg->cdata.hint = get32n(ptr);
    msg->cdata.data = ptr + 4;
    msg->cdata.len = len - 4;

    return true;
}

int encode_cdata_message(const struct pktlab_message *msg, struct iovec *iov) {
    trace("%s(msg:{hint:%d})", __func__, msg->cdata.hint);
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;

    set32n(ptr, msg->cdata.hint);
    iov[0].iov_len += 4;

    if (msg->cdata.len) {
        iov[1].iov_base = (void *)msg->cdata.data;
        iov[1].iov_len = msg->cdata.len;
        return 2;
    }
    return 1;
}

bool decode_start_message(struct pktlab_message *msg, const void *ptr,
                          uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    msg->type = PKTLAB_START_MESSAGE;
    return (len == 0);
}

int encode_start_message(const struct pktlab_message *msg, struct iovec *iov) {
    trace("%s(msg)", __func__);
    return 1;
}

bool decode_end_message(struct pktlab_message *msg, const void *ptr,
                        uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    msg->type = PKTLAB_END_MESSAGE;
    return (len == 0);
}

int encode_end_message(const struct pktlab_message *msg, struct iovec *iov) {
    trace("%s(msg)", __func__);
    return 1;
}

bool decode_yield_message(struct pktlab_message *msg, const void *ptr,
                          uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);

    RETURN_IF(len != 8, false);
    msg->type = PKTLAB_YIELD_MESSAGE;
    msg->yield.rhint = get64n(ptr);
    return true;
}

int encode_yield_message(const struct pktlab_message *msg, struct iovec *iov) {
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;

    trace("%s(msg:{rhint:%" PRIu64 "})", __func__, msg->yield.rhint);
    set64n(ptr, msg->yield.rhint);
    iov[0].iov_len += 8;
    return 1;
}

bool decode_cont_message(struct pktlab_message *msg, const void *ptr,
                         uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    msg->type = PKTLAB_CONT_MESSAGE;
    return (len == 0);
}

int encode_cont_message(const struct pktlab_message *msg, struct iovec *iov) {
    trace("%s(msg)", __func__);
    return 1;
}

bool decode_mread_message(struct pktlab_message *msg, const void *ptr,
                          uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    RETURN_IF(len != 6, false);

    msg->type = PKTLAB_MREAD_MESSAGE;
    msg->mread.addr = get32n(ptr);
    msg->mread.len = get16n(ptr + 4);
    return (msg->mread.len > 0) &&
           (msg->mread.addr <= UINT32_MAX - msg->mread.len);
}

int encode_mread_message(const struct pktlab_message *msg, struct iovec *iov) {
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;

    trace("%s(msg:{addr:%u,len:%u})", __func__, (unsigned int)msg->mread.addr,
          (unsigned int)msg->mread.len);

    set32n(ptr + 0, msg->mread.addr);
    set16n(ptr + 4, msg->mread.len);
    iov[0].iov_len += 6;
    return 1;
}

bool decode_mwrite_message(struct pktlab_message *msg, const void *ptr,
                           uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    RETURN_IF(len < 4, false);

    msg->type = PKTLAB_MWRITE_MESSAGE;
    msg->mwrite.addr = get32n(ptr);
    msg->mwrite.ptr = ptr + 4;
    msg->mwrite.len = len - 4;
    return (msg->mwrite.addr <= UINT32_MAX - msg->mwrite.len);
}

int encode_mwrite_message(const struct pktlab_message *msg, struct iovec *iov) {
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;
    trace("%s(msg:{addr:%u,len:%u})", __func__, (unsigned int)msg->mwrite.addr,
          (unsigned int)msg->mwrite.len);
    set32n(ptr, msg->mwrite.addr);
    iov[0].iov_len += 4;

    iov[1].iov_base = (void *)msg->mwrite.ptr;
    iov[1].iov_len = msg->mwrite.len;
    return 2;
}

bool decode_nopen_message(struct pktlab_message *msg, const void *ptr,
                          uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    RETURN_IF(len < 7, false);

    msg->type = PKTLAB_NOPEN_MESSAGE;
    msg->nopen.sktid = get8(ptr);
    msg->nopen.family = get8(ptr + 1);
    RETURN_IF(!(msg->nopen.family == PKTLAB_IP4_PROTO ||
                msg->nopen.family == PKTLAB_IP6_PROTO),
              false);
    msg->nopen.proto = get8(ptr + 2);
    msg->nopen.rbufsz = get32n(ptr + 3);

    switch (msg->nopen.proto) {
    case PKTLAB_RAW_PROTO:
        msg->nopen.laddrptr = ptr + 7;
        msg->nopen.laddrlen = len - 7;
        if (msg->nopen.family == PKTLAB_IP4_PROTO) {
            return (msg->nopen.laddrlen == 4);
        } else {
            return (msg->nopen.laddrlen == 16);
        }
    case PKTLAB_UDP_PROTO:
        msg->nopen.laddrptr = ptr + 7;
        if (msg->nopen.family == PKTLAB_IP4_PROTO) {
            msg->nopen.laddrlen = 4;
            RETURN_IF(len != 13, false);
        } else {
            msg->nopen.laddrlen = 16;
            RETURN_IF(len != 25, false);
        }
        msg->nopen.lportptr = ptr + 7 + msg->nopen.laddrlen;
        msg->nopen.lportlen = 2;
        break;
    case PKTLAB_TCP_PROTO:
        if (msg->nopen.family == PKTLAB_IP4_PROTO) {
            if (len - 7 == 12) {
                // tcp_connect
                // msg->nopen.nopen_type = PKTLAB_NOPEN_TYPE_TCP_CONNECT;
                msg->nopen.laddrptr = ptr + 7;
                msg->nopen.laddrlen = 4;
                msg->nopen.lportptr = ptr + 7 + 4;
                msg->nopen.lportlen = 2;
                msg->nopen.raddrptr = ptr + 7 + 6;
                msg->nopen.raddrlen = 4;
                msg->nopen.rportptr = ptr + 7 + 6 + 4;
                msg->nopen.rportlen = 2;
            } else if (len - 7 == 6) {
                // tcp_listen
                // msg->nopen.nopen_type = PKTLAB_NOPEN_TYPE_TCP_LISTEN;
                msg->nopen.laddrptr = ptr + 7;
                msg->nopen.laddrlen = 4;
                msg->nopen.lportptr = ptr + 7 + 4;
                msg->nopen.lportlen = 2;
                msg->nopen.raddrptr = NULL;
                msg->nopen.raddrlen = 0;
                msg->nopen.rportptr = NULL;
                msg->nopen.rportlen = 0;
            }
        } else {
            if (len - 7 == 36) {
                // tcp_connect
                // msg->nopen.nopen_type = PKTLAB_NOPEN_TYPE_TCP_CONNECT;
                msg->nopen.laddrptr = ptr + 7;
                msg->nopen.laddrlen = 16;
                msg->nopen.lportptr = ptr + 7 + 16;
                msg->nopen.lportlen = 2;
                msg->nopen.raddrptr = ptr + 7 + 18;
                msg->nopen.raddrlen = 16;
                msg->nopen.rportptr = ptr + 7 + 18 + 16;
                msg->nopen.rportlen = 2;
            } else if (len - 7 == 18) {
                // tcp_listen
                // msg->nopen.nopen_type = PKTLAB_NOPEN_TYPE_TCP_LISTEN;
                msg->nopen.laddrptr = ptr + 7;
                msg->nopen.laddrlen = 16;
                msg->nopen.lportptr = ptr + 7 + 16;
                msg->nopen.lportlen = 2;
                msg->nopen.raddrptr = NULL;
                msg->nopen.raddrlen = 0;
                msg->nopen.rportptr = NULL;
                msg->nopen.rportlen = 0;
            }
        }
        break;
    default:
        return false;
    }
    return true;
}

int encode_nopen_message(const struct pktlab_message *msg, struct iovec *iov) {
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;
    trace("%s(msg:{id:%d,pf:%u,pr:%u})", __func__,
          (unsigned int)msg->nopen.sktid, (unsigned int)msg->nopen.family,
          (unsigned int)msg->nopen.proto);

    set8(ptr, msg->nopen.sktid);
    set8(ptr + 1, msg->nopen.family);
    set8(ptr + 2, msg->nopen.proto);
    set32n(ptr + 3, msg->nopen.rbufsz);

    iov[0].iov_len += 7;

    // laddr requirement applies for all types of socket
    if ((msg->nopen.family == PKTLAB_IP4_PROTO &&
         msg->nopen.laddrlen != PKTLAB_IPV4_WO_MSK_ADDR_LEN) ||
        (msg->nopen.family == PKTLAB_IP6_PROTO &&
         msg->nopen.laddrlen != PKTLAB_IPV6_ADDR_LEN))
        return -1;

    switch (msg->nopen.proto) {
    case PKTLAB_RAW_PROTO:
        iov[1].iov_base = (void *)msg->nopen.laddrptr;
        iov[1].iov_len = msg->nopen.laddrlen;
        return 2;
    case PKTLAB_UDP_PROTO:
        iov[1].iov_base = (void *)msg->nopen.laddrptr;
        iov[1].iov_len = msg->nopen.laddrlen;
        iov[2].iov_base = (void *)msg->nopen.lportptr;
        iov[2].iov_len = msg->nopen.lportlen;
        return 3;
    case PKTLAB_TCP_PROTO:
        // TCP listen
        if (msg->nopen.raddrlen == 0 || msg->nopen.raddrptr == NULL ||
            msg->nopen.rportlen == 0 || msg->nopen.rportptr == NULL) {
            debug("nopen tcp listen");
            iov[1].iov_base = (void *)msg->nopen.laddrptr;
            iov[1].iov_len = msg->nopen.laddrlen;
            iov[2].iov_base = (void *)msg->nopen.lportptr;
            iov[2].iov_len = msg->nopen.lportlen;
            return 3;
        } else {
            debug("nopen tcp connect");
            if (msg->nopen.family == PKTLAB_IP4_PROTO &&
                msg->nopen.raddrlen != 4)
                return -1;
            if (msg->nopen.family == PKTLAB_IP6_PROTO &&
                msg->nopen.raddrlen != 16)
                return -1;
            iov[1].iov_base = (void *)msg->nopen.laddrptr;
            iov[1].iov_len = msg->nopen.laddrlen;
            iov[2].iov_base = (void *)msg->nopen.lportptr;
            iov[2].iov_len = msg->nopen.lportlen;
            iov[3].iov_base = (void *)msg->nopen.raddrptr;
            iov[3].iov_len = msg->nopen.raddrlen;
            iov[4].iov_base = (void *)msg->nopen.rportptr;
            iov[4].iov_len = msg->nopen.rportlen;
            return 5;
        }
    default:
        return -1;
    }
}
bool decode_nclose_message(struct pktlab_message *msg, const void *ptr,
                           uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);

    RETURN_IF(len != 1, false);

    msg->type = PKTLAB_NCLOSE_MESSAGE;
    msg->nclose.sktid = get8(ptr);
    return true;
}

int encode_nclose_message(const struct pktlab_message *msg, struct iovec *iov) {
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;

    trace("%s(msg:{sktid:%d})", __func__, (int)msg->nclose.sktid);

    set8(ptr + 0, (uint8_t)msg->nclose.sktid);
    iov[0].iov_len += 1;
    return 1;
}
bool decode_nsend_message(struct pktlab_message *msg, const void *ptr,
                          uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    RETURN_IF(len < 13, false);

    msg->type = PKTLAB_NSEND_MESSAGE;
    msg->nsend.sktid = get8(ptr);
    msg->nsend.family = get8(ptr + 1);
    msg->nsend.proto = get8(ptr + 2);
    msg->nsend.time = get64n(ptr + 3);
    msg->nsend.tag = get16n(ptr + 11);
    switch (msg->nsend.proto) {
    case PKTLAB_RAW_PROTO:
    case PKTLAB_TCP_PROTO:
        msg->nsend.ptr = ptr + 13;
        msg->nsend.len = len - 13;
        break;
    case PKTLAB_UDP_PROTO:
        msg->nsend.udp.raddrptr = ptr + 13;
        if (msg->nsend.family == PKTLAB_IP4_PROTO) {
            msg->nsend.udp.raddrlen = 4;
        } else if (msg->nsend.family == PKTLAB_IP6_PROTO) {
            msg->nsend.udp.raddrlen = 16;
        } else {
            return false;
        }
        msg->nsend.udp.rportptr = ptr + 13 + msg->nsend.udp.raddrlen;
        msg->nsend.udp.rportlen = 2;
        msg->nsend.ptr = ptr + 13 + msg->nsend.udp.raddrlen + 2;
        msg->nsend.len = len - (13 + msg->nsend.udp.raddrlen + 2);
        break;

    default:
        return false;
    }
    return true;
}

int encode_nsend_message(const struct pktlab_message *msg, struct iovec *iov) {
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;
    trace("%s(msg:{sktid:%d})", __func__, (int)msg->nsend.sktid);
    set8(ptr, msg->nsend.sktid);
    set8(ptr + 1, msg->nsend.family);
    set8(ptr + 2, msg->nsend.proto);
    set64n(ptr + 3, msg->nsend.time);
    set16n(ptr + 11, msg->nsend.tag);
    iov[0].iov_len += 13;

    switch (msg->nsend.proto) {
    case PKTLAB_RAW_PROTO:
    case PKTLAB_TCP_PROTO:
        iov[1].iov_base = (void *)msg->nsend.ptr;
        iov[1].iov_len = msg->nsend.len;
        return 2;
    case PKTLAB_UDP_PROTO:
        iov[1].iov_base = (void *)msg->nsend.udp.raddrptr;
        if (msg->nsend.family == PKTLAB_IP4_PROTO) {
            iov[1].iov_len = 4;
        } else if (msg->nsend.family == PKTLAB_IP6_PROTO) {
            iov[1].iov_len = 16;
        } else {
            return -1;
        }
        iov[2].iov_base = (void *)msg->nsend.udp.rportptr;
        iov[2].iov_len = msg->nsend.udp.rportlen;
        iov[3].iov_base = (void *)msg->nsend.ptr;
        iov[3].iov_len = msg->nsend.len;
        return 4;
    default:
        break;
    }
    return -1;
}

bool decode_ncap_message(struct pktlab_message *msg, const void *ptr,
                         uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);

    RETURN_IF(len < 12, false);

    msg->type = PKTLAB_NCAP_MESSAGE;
    msg->ncap.sktid = get8(ptr);
    msg->ncap.family = get8(ptr + 1);
    msg->ncap.proto = get8(ptr + 2);
    msg->ncap.endtime = get64n(ptr + 3);
    msg->ncap.raw.ftype = get8(ptr + 11);
    msg->ncap.raw.filter = ptr + 12;
    msg->ncap.raw.filterlen = len - 12; // 0 len filter considered well-formed
    return true;
}

int encode_ncap_message(const struct pktlab_message *msg, struct iovec *iov) {
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;
    int iovcnt = 1;

    trace("%s(msg:{sktid:%d,ftype:%d,filterlen:%u})", __func__,
          (int)msg->ncap.sktid, (int)msg->ncap.raw.ftype,
          (unsigned)msg->ncap.raw.filterlen);

    RETURN_IF((msg->ncap.raw.filter == NULL) != (msg->ncap.raw.filterlen == 0),
              -1);

    set8(ptr, msg->ncap.sktid);
    set8(ptr + 1, msg->ncap.family);
    set8(ptr + 2, msg->ncap.proto);
    set64n(ptr + 3, msg->ncap.endtime);
    set8(ptr + 11, msg->ncap.raw.ftype);
    iov[0].iov_len += 12;

    if (msg->ncap.raw.filterlen != 0) {
        iov[1].iov_base = (void *)msg->ncap.raw.filter;
        iov[1].iov_len = msg->ncap.raw.filterlen;
        ++iovcnt;
    }

    return iovcnt;
}
bool decode_nctl_message(struct pktlab_message *msg, const void *ptr,
                         uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    RETURN_IF(len < 11, false);
    msg->type = PKTLAB_NCTL_MESSAGE;
    msg->nctl.sktid = get8(ptr);
    msg->nctl.family = get8(ptr + 1);
    msg->nctl.proto = get8(ptr + 2);
    msg->nctl.time = get64n(ptr + 3);
    switch (msg->nctl.proto) {
    case PKTLAB_TCP_PROTO:
        msg->nctl.tcp.command = get8(ptr + 11);
        break;

    default:
        return false;
    }
    return true;
}

int encode_nctl_message(const struct pktlab_message *msg, struct iovec *iov) {
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;
    trace("%s(msg:{sktid:%d})", __func__, (int)msg->nctl.sktid);
    set8(ptr, msg->nctl.sktid);
    set8(ptr + 1, msg->nctl.family);
    set8(ptr + 2, msg->nctl.proto);
    set64n(ptr + 3, msg->nctl.time);
    iov[0].iov_len += 11;
    switch (msg->nctl.proto) {
    case PKTLAB_TCP_PROTO:
        set8(ptr + 11, msg->nctl.tcp.command);
        iov[0].iov_len += 1;
        return 1;

    default:
        return -1;
        break;
    }
    return -1;
}
bool decode_mdata_message(struct pktlab_message *msg, const void *ptr,
                          uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    RETURN_IF(len < 4, false);
    msg->type = PKTLAB_MDATA_MESSAGE;
    msg->mdata.addr = get32n(ptr);
    msg->mdata.ptr = ptr + 4;
    msg->mdata.len = len - 4;
    return true;
}

int encode_mdata_message(const struct pktlab_message *msg, struct iovec *iov) {
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;
    trace("%s(msg:{addr:%d})", __func__, (int)msg->mdata.addr);
    set32n(ptr, msg->mdata.addr);
    iov[0].iov_len += 4;

    iov[1].iov_base = (void *)msg->mdata.ptr;
    iov[1].iov_len = msg->mdata.len;
    return 2;
}

bool decode_ntag_message(struct pktlab_message *msg, const void *ptr,
                         uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    RETURN_IF(len < 11, false);
    msg->type = PKTLAB_NTAG_MESSAGE;
    msg->ntag.sktid = get8(ptr);
    msg->ntag.tag = get16n(ptr + 1);
    msg->ntag.time = get64n(ptr + 3);

    if (len == 11) {
        msg->ntag.errid = PKTLAB_SUCCESS;
        return true;
    }
    if (len == 13) {
        msg->ntag.errid = get16n(ptr + 11);
        return true;
    }
    return false;
}

int encode_ntag_message(const struct pktlab_message *msg, struct iovec *iov) {
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;
    trace("%s(msg:{sktid:%d})", __func__, (int)msg->ntag.sktid);
    set8(ptr, msg->ntag.sktid);
    set16n(ptr + 1, msg->ntag.tag);
    set64n(ptr + 3, msg->ntag.time);
    iov[0].iov_len += 11;

    if (msg->ntag.errid != PKTLAB_SUCCESS) {
        set16n(ptr + 11, msg->ntag.errid);
        iov[0].iov_len += 2;
    }

    return 1;
}

bool decode_nstat_message(struct pktlab_message *msg, const void *ptr,
                          uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    RETURN_IF(len < 12, false);
    msg->type = PKTLAB_NSTAT_MESSAGE;
    msg->nstat.sktid = get8(ptr);
    msg->nstat.family = get8(ptr + 1);
    msg->nstat.proto = get8(ptr + 2);
    msg->nstat.time = get64n(ptr + 3);

    if (len == 12) {
        msg->nstat.code = get8(ptr + 11);
        return true;
    }

    if (len == 17 || len == 29) {
        msg->nstat.code = PKTLAB_TCPEV_ESTABLISHED;
        msg->nstat.tcp_established.raddrptr = ptr + 11;
        msg->nstat.tcp_established.raddrlen = len - 11 - 2;
        msg->nstat.tcp_established.rportptr = ptr + len - 2;
        msg->nstat.tcp_established.rportlen = 2;
        return true;
    }
    return false;
}

int encode_nstat_message(const struct pktlab_message *msg, struct iovec *iov) {
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;
    trace("%s(msg:{sktid:%d,code:%d})", __func__, (int)msg->nstat.sktid,
          msg->nstat.code);
    set8(ptr, msg->nstat.sktid);
    set8(ptr + 1, msg->nstat.family);
    set8(ptr + 2, msg->nstat.proto);
    set64n(ptr + 3, msg->nstat.time);
    iov[0].iov_len += 11;

    if (msg->nstat.proto == PKTLAB_TCP_PROTO &&
        msg->nstat.code == PKTLAB_TCPEV_ESTABLISHED) {
        switch (msg->nstat.family) {
        case PKTLAB_IP4_PROTO:
            iov[1].iov_base = (void *)msg->nstat.tcp_established.raddrptr;
            iov[1].iov_len = 4;
            iov[2].iov_base = (void *)msg->nstat.tcp_established.rportptr;
            iov[2].iov_len = 2;
            return 3;
        case PKTLAB_IP6_PROTO:
            iov[1].iov_base = (void *)msg->nstat.tcp_established.raddrptr;
            iov[1].iov_len = 16;
            iov[2].iov_base = (void *)msg->nstat.tcp_established.rportptr;
            iov[2].iov_len = 2;
            return 3;
        default:
            return -1;
        }
    } else {
        set8(ptr + 11, msg->nstat.code);
        iov[0].iov_len += 1;
    }
    return 1;
}

bool decode_ndata_message(struct pktlab_message *msg, const void *ptr,
                          uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    RETURN_IF(len < 11, false);
    msg->type = PKTLAB_NDATA_MESSAGE;
    msg->ndata.sktid = get8(ptr);
    msg->ndata.family = get8(ptr + 1);
    msg->ndata.proto = get8(ptr + 2);
    msg->ndata.time = get64n(ptr + 3);

    switch (msg->ndata.proto) {
    case PKTLAB_RAW_PROTO:
    case PKTLAB_TCP_PROTO:
        msg->ndata.ptr = ptr + 11;
        msg->ndata.len = len - 11;
        break;
    case PKTLAB_UDP_PROTO:
        if (msg->ndata.family == PKTLAB_IP4_PROTO) {
            msg->ndata.udp.raddrptr = ptr + 11;
            msg->ndata.udp.raddrlen = 4;
            msg->ndata.udp.rportptr = ptr + 11 + 4;
            msg->ndata.udp.rportlen = 2;
        } else if (msg->ndata.family == PKTLAB_IP6_PROTO) {
            msg->ndata.udp.raddrptr = ptr + 11;
            msg->ndata.udp.raddrlen = 16;
            msg->ndata.udp.rportptr = ptr + 11 + 16;
            msg->ndata.udp.rportlen = 2;
        } else {
            return false;
        }
        msg->ndata.ptr = ptr + 11 + msg->ndata.udp.raddrlen + 2;
        msg->ndata.len = ptr + len - msg->ndata.ptr;
        break;

    default:
        return false;
    }
    return true;
}

int encode_ndata_message(const struct pktlab_message *msg, struct iovec *iov) {
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;
    trace("%s(msg:{sktid:%d})", __func__, (int)msg->ndata.sktid);
    set8(ptr, msg->ndata.sktid);
    set8(ptr + 1, msg->ndata.family);
    set8(ptr + 2, msg->ndata.proto);
    set64n(ptr + 3, msg->ndata.time);
    iov[0].iov_len += 11;

    switch (msg->ndata.proto) {
    case PKTLAB_RAW_PROTO:
    case PKTLAB_TCP_PROTO:
        iov[1].iov_base = (void *)msg->ndata.ptr;
        iov[1].iov_len = msg->ndata.len;
        return 2;
    case PKTLAB_UDP_PROTO:
        if (msg->ndata.family == PKTLAB_IP4_PROTO ||
            msg->ndata.family == PKTLAB_IP6_PROTO) {
            iov[1].iov_base = (void *)msg->ndata.udp.raddrptr;
            iov[1].iov_len = msg->ndata.udp.raddrlen;
            iov[2].iov_base = (void *)msg->ndata.udp.rportptr;
            iov[2].iov_len = msg->ndata.udp.rportlen;
            iov[3].iov_base = (void *)msg->ndata.ptr;
            iov[3].iov_len = msg->ndata.len;
            return 4;
        }
    default:
        break;
    }
    return -1;
}

bool decode_ndrop_message(struct pktlab_message *msg, const void *ptr,
                          uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    RETURN_IF(len != 27, false);
    msg->type = PKTLAB_NDROP_MESSAGE;
    msg->ndrop.sktid = get8(ptr);
    msg->ndrop.family = get8(ptr + 1);
    msg->ndrop.proto = get8(ptr + 2);

    ptr += 3;
    msg->ndrop.tfirst = get64n(ptr);
    msg->ndrop.tlast = get64n(ptr + 8);

    ptr += 16;
    msg->ndrop.unitcnt = get32n(ptr);
    msg->ndrop.bytecnt = get32n(ptr + 4);
    return true;
}

int encode_ndrop_message(const struct pktlab_message *msg, struct iovec *iov) {
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;
    trace("%s(msg:{sktid:%d})", __func__, (int)msg->ndrop.sktid);
    set8(ptr, msg->ndrop.sktid);
    set8(ptr + 1, msg->ndrop.family);
    set8(ptr + 2, msg->ndrop.proto);
    ptr += 3;
    set64n(ptr, msg->ndrop.tfirst);
    set64n(ptr + 8, msg->ndrop.tlast);
    ptr += 16;
    set32n(ptr, msg->ndrop.unitcnt);
    set32n(ptr + 4, msg->ndrop.bytecnt);
    iov[0].iov_len += 27;
    return 1;
}

bool decode_xcert_message(struct pktlab_message *msg, const void *ptr,
                          uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);

    RETURN_IF(len == 0, false);

    // we don't check if the der is well-formed here
    msg->type = PKTLAB_XCERT_MESSAGE;
    msg->xcert.derptr = ptr;
    msg->xcert.derlen = len;
    return true;
}

int encode_xcert_message(const struct pktlab_message *msg, struct iovec *iov) {
    trace("%s(msg:{derlen:%" PRIu32 "})", __func__, msg->xcert.derlen);

    RETURN_IF(msg->xcert.derptr == NULL || msg->xcert.derlen == 0, -1);

    iov[1].iov_base = (void *)msg->xcert.derptr;
    iov[1].iov_len = msg->xcert.derlen;
    return 2;
}

bool decode_xpub_message(struct pktlab_message *msg, const void *ptr,
                         uint_fast32_t len) {
    const void *chlistend;
    uint_fast32_t chlistlen;

    trace("%s(len:%" PRIuFAST32 ")", __func__, len);

    RETURN_IF(len < PKTLAB_PUBLICKEY_LEN + 1, false);

    msg->type = PKTLAB_XPUB_MESSAGE;

    RETURN_IF(find_chlistend(ptr, len, &chlistend, &chlistlen) < 0, false);
    RETURN_IF(len < chlistlen + PKTLAB_PUBLICKEY_LEN, false);
    assert(ptr + chlistlen == chlistend);

    msg->xpub.chlistptr = ptr;
    msg->xpub.chlistlen = chlistlen;
    msg->xpub.eopkey = chlistend;
    msg->xpub.xdescrptr = NULL;
    msg->xpub.xdescrlen = 0;

    if (len > chlistlen + PKTLAB_PUBLICKEY_LEN) {
        msg->xpub.xdescrptr = ptr + chlistlen + PKTLAB_PUBLICKEY_LEN;
        msg->xpub.xdescrlen = len - chlistlen - PKTLAB_PUBLICKEY_LEN;

        RETURN_IF(!is_good_xdescr(msg->xpub.xdescrptr, msg->xpub.xdescrlen),
                  false);
    }

    return true;
}

int encode_xpub_message(const struct pktlab_message *msg, struct iovec *iov) {
    int cur_iov = 1;

    trace("%s(msg:{chlistlen:%" PRIu32 ",xdescrlen:%" PRIu32 "})", __func__,
          msg->xpub.chlistlen, msg->xpub.xdescrlen);

    RETURN_IF(!is_good_chlist(msg->xpub.chlistptr, msg->xpub.chlistlen) ||
                  msg->xpub.eopkey == NULL ||
                  !is_good_xdescr(msg->xpub.xdescrptr, msg->xpub.xdescrlen),
              -1);

    iov[cur_iov].iov_base = (void *)msg->xpub.chlistptr;
    iov[cur_iov].iov_len = msg->xpub.chlistlen;
    ++cur_iov;

    iov[cur_iov].iov_base = (void *)msg->xpub.eopkey;
    iov[cur_iov].iov_len = PKTLAB_PUBLICKEY_LEN;
    ++cur_iov;

    if (msg->xpub.xdescrlen != 0) {
        iov[cur_iov].iov_base = (void *)msg->xpub.xdescrptr;
        iov[cur_iov].iov_len = msg->xpub.xdescrlen;
        ++cur_iov;
    }

    return cur_iov;
}

bool decode_xsub_message(struct pktlab_message *msg, const void *ptr,
                         uint_fast32_t len) {
    const void *chlistend;
    uint_fast32_t chlistlen;

    trace("%s(len:%" PRIuFAST32 ")", __func__, len);

    RETURN_IF(len < PKTLAB_PUBLICKEY_LEN + 1, false);

    msg->type = PKTLAB_XSUB_MESSAGE;

    RETURN_IF(find_chlistend(ptr, len, &chlistend, &chlistlen) < 0, false);
    RETURN_IF(len < chlistlen + PKTLAB_PUBLICKEY_LEN, false);
    assert(ptr + chlistlen == chlistend);

    msg->xsub.chlistptr = ptr;
    msg->xsub.chlistlen = chlistlen;
    msg->xsub.eopkey = chlistend;
    msg->xsub.dattrptr = NULL;
    msg->xsub.dattrlen = 0;

    if (len > chlistlen + PKTLAB_PUBLICKEY_LEN) {
        msg->xsub.dattrptr = ptr + chlistlen + PKTLAB_PUBLICKEY_LEN;
        msg->xsub.dattrlen = len - chlistlen - PKTLAB_PUBLICKEY_LEN;

        RETURN_IF(!is_good_dattr(msg->xsub.dattrptr, msg->xsub.dattrlen),
                  false);
    }

    return true;
}

int encode_xsub_message(const struct pktlab_message *msg, struct iovec *iov) {
    int cur_iov = 1;

    trace("%s(msg:{chlistlen:%" PRIu32 ",dattrlen:%" PRIu32 "})", __func__,
          msg->xsub.chlistlen, msg->xsub.dattrlen);

    RETURN_IF(!is_good_chlist(msg->xsub.chlistptr, msg->xsub.chlistlen) ||
                  msg->xsub.eopkey == NULL ||
                  !is_good_dattr(msg->xsub.dattrptr, msg->xsub.dattrlen),
              -1);

    iov[cur_iov].iov_base = (void *)msg->xsub.chlistptr;
    iov[cur_iov].iov_len = msg->xsub.chlistlen;
    ++cur_iov;

    iov[cur_iov].iov_base = (void *)msg->xsub.eopkey;
    iov[cur_iov].iov_len = PKTLAB_PUBLICKEY_LEN;
    ++cur_iov;

    if (msg->xsub.dattrlen != 0) {
        iov[cur_iov].iov_base = (void *)msg->xsub.dattrptr;
        iov[cur_iov].iov_len = msg->xsub.dattrlen;
        ++cur_iov;
    }

    return cur_iov;
}

bool decode_suspd_message(struct pktlab_message *msg, const void *ptr,
                          uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    msg->type = PKTLAB_SUSPD_MESSAGE;
    return (len == 0);
}

int encode_suspd_message(const struct pktlab_message *msg, struct iovec *iov) {
    trace("%s(msg)", __func__);
    return 1;
}

bool decode_resumd_message(struct pktlab_message *msg, const void *ptr,
                           uint_fast32_t len) {
    trace("%s(len:%" PRIuFAST32 ")", __func__, len);
    msg->type = PKTLAB_RESUMD_MESSAGE;
    return (len == 0);
}

int encode_resumd_message(const struct pktlab_message *msg, struct iovec *iov) {
    trace("%s(msg)", __func__);
    return 1;
}

bool decode_xnotify_message(struct pktlab_message *msg, const void *ptr,
                            uint_fast32_t len) {

    trace("%s(len:%" PRIuFAST32 ")", __func__, len);

    RETURN_IF(len == 0, false);

    msg->type = PKTLAB_XNOTIFY_MESSAGE;
    msg->xnotify.channellen = get8(ptr);
    msg->xnotify.channelptr = ptr + 1;
    msg->xnotify.xdescrptr = NULL;
    msg->xnotify.xdescrlen = 0;

    RETURN_IF(msg->xnotify.channellen == 0 ||
                  msg->xnotify.channellen > PKTLAB_CHANNELID_LEN_MAX ||
                  len < msg->xnotify.channellen + 1,
              false);

    if (len > msg->xnotify.channellen + 1) {
        msg->xnotify.xdescrptr = ptr + msg->xnotify.channellen + 1;
        msg->xnotify.xdescrlen = len - msg->xnotify.channellen - 1;

        RETURN_IF(
            !is_good_xdescr(msg->xnotify.xdescrptr, msg->xnotify.xdescrlen),
            false);
    }

    return true;
}

int encode_xnotify_message(const struct pktlab_message *msg,
                           struct iovec *iov) {
    uint8_t *ptr = iov[0].iov_base + iov[0].iov_len;

    trace("%s(msg:{channellen:%" PRIu8 ",xdescrlen:%" PRIu32 "})", __func__,
          msg->xnotify.channellen, msg->xnotify.xdescrlen);

    RETURN_IF(
        msg->xnotify.channelptr == NULL || msg->xnotify.channellen == 0 ||
            msg->xnotify.channellen > PKTLAB_CHANNELID_LEN_MAX ||
            !is_good_xdescr(msg->xnotify.xdescrptr, msg->xnotify.xdescrlen),
        -1);

    set8(ptr, msg->xnotify.channellen);
    ++iov[0].iov_len;

    iov[1].iov_base = (void *)msg->xnotify.channelptr;
    iov[1].iov_len = msg->xnotify.channellen;

    RETURN_IF(msg->xnotify.xdescrlen == 0, 2);

    iov[2].iov_base = (void *)msg->xnotify.xdescrptr;
    iov[2].iov_len = msg->xnotify.xdescrlen;
    return 3;
}

bool decode_xreg_message(struct pktlab_message *msg, const void *ptr,
                         uint_fast32_t len) {

    trace("%s(len:%" PRIuFAST32 ")", __func__, len);

    msg->type = PKTLAB_XREG_MESSAGE;
    msg->xreg.snikeyptr = ptr;
    msg->xreg.snikeylen = len;

    return is_good_snikey(msg->xreg.snikeyptr, msg->xreg.snikeylen);
}

int encode_xreg_message(const struct pktlab_message *msg, struct iovec *iov) {

    trace("%s(msg:{xdescrlen:%" PRIu32 "})", __func__, msg->xnotify.xdescrlen);

    RETURN_IF(!is_good_snikey(msg->xreg.snikeyptr, msg->xreg.snikeylen), -1);

    iov[1].iov_base = (void *)msg->xreg.snikeyptr;
    iov[1].iov_len = msg->xreg.snikeylen;
    return 2;
}

bool is_good_xdescr(const void *ptr, uint_fast32_t len) {
    struct pktlab_xdescr_dattr_pair *pairs;
    uint_fast32_t paircnt;

    if (pktlab_decode_xdescr(ptr, len, &pairs, &paircnt) < 0)
        return false;

    free(pairs);
    return true;
}

bool is_good_dattr(const void *ptr, uint_fast32_t len) {
    struct pktlab_xdescr_dattr_pair *pairs;
    uint_fast32_t paircnt;

    if (pktlab_decode_dattr(ptr, len, &pairs, &paircnt) < 0)
        return false;

    free(pairs);
    return true;
}

bool is_good_chlist(const void *ptr, uint_fast32_t len) {
    struct pktlab_channel *channels;

    if (pktlab_decode_chlist(ptr, len, &channels) < 0)
        return false;

    pktlab_free_channels(channels);
    return true;
}

bool is_good_snikey(const void *ptr, uint_fast32_t len) {
    char buf[PKTCTRL_SNIKEY_LEN_MAX + 1] = {};

    if (len > PKTCTRL_SNIKEY_LEN_MAX)
        return false;

    memcpy(buf, ptr, len);

    return pktctrl_is_valid_snikey(buf);
}