// main.c
//

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pthread.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/time.h>
#include <unistd.h>

#include "console.h"
#include "manager.h"
#include "util.h"

//
// INTERNAL CONSTANTS
//

#define USAGE "pktxpmgr [-h] [-c XM_CONF_PATH] ADDR[:PORT] MLET [MLET_ARGS ...]"
#define DEFAULT_ECCONF_PATH "/.pktlab/xpmgr.conf" // prepended with HOME env var
#define FALLBACK_ECCONF_PATH "xpmgr.conf"

#define MAX_SELECT_SEC 2678400 // 31 days in seconds

#define FILEPATH_BUFSZ 4096

//
// INTERNAL TYPE DEFINITIONS
//

enum term_st {
    RUNNING = 0,
    WRAPUP  = 1,
    END     = 2
};

//
// INTERNAL GLOBAL VARIABLES
//

static bool _stdinout_closed;

static pthread_mutex_t _st_lock = PTHREAD_MUTEX_INITIALIZER;
static int _st;

//
// INTERNAL FUNCTION DECLARATIONS
//

static int signal_handling_setup(void);
static void sigint_handler(int signal);

static void close_stdinout(void);

static void run_manager (
    const char * confpath,
    const char * addr,
    char * exp_prog_n_args[]);

static void get_st(int * cur_st);

//
// MAIN FUNCTION DEFINITION
//

int main(int argc, char * argv[]) {
    int optch, rst;
    const char * homepath;
    size_t homepathlen;
    char confpath_buf[FILEPATH_BUFSZ];
    const char * confpath = NULL;
    const char * addr;
    char ** exp_prog_n_args;

    // Parse command-line arguments
    while ((optch = getopt(argc, argv, "c:h")) != -1) {
        switch (optch) {
        case 'c':
            confpath = optarg;
            break;
        case 'h':
        default:
            goto usage_error;
        }
    }

    argc -= optind;
    argv += optind;

    // set up default option/argument values
    if (argc < 2)
        goto usage_error;

    if (confpath == NULL) {
        homepath = getenv("HOME");
        if (homepath == NULL) {
            warn("Cannot get HOME environment variable value");
            warn("Looking for %s in current working directory instead",
                FALLBACK_ECCONF_PATH);
            confpath = FALLBACK_ECCONF_PATH;
        } else {
            homepathlen = strlen(homepath);
            if (homepathlen > sizeof(confpath_buf)-sizeof(DEFAULT_ECCONF_PATH)) {
                fatal("Config path buffer overflow: HOME path too long (%zu)",
                    homepathlen);
                exit(EXIT_FAILURE);
            }

            strncpy(confpath_buf, homepath, sizeof(confpath_buf));
            strncpy(
                confpath_buf+homepathlen, DEFAULT_ECCONF_PATH,
                sizeof(confpath_buf)-homepathlen); // '\0' appended
            confpath = confpath_buf;
        }
    }

    addr = argv[0];
    exp_prog_n_args = argv+1;

    // set up environment and run manager
    rst = signal_handling_setup();
    assert(rst == 0);
    close_stdinout();

    info("Running manager");
    info("Config path:\t%s", safe_str(confpath));
    info("Commandline address:\t%s", safe_str(addr));
    info("Experiment program path:\t%s", safe_str(exp_prog_n_args[0]));

    run_manager(confpath, addr, exp_prog_n_args);
    return 0;

usage_error:
    fprintf(stderr, "USAGE: " USAGE "\n");
    exit(EXIT_FAILURE);
    return -1;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

int signal_handling_setup(void) {
    struct sigaction action;

    // Disable SIGPIPE signal; handle broken pipe error on write.
    memset(&action, 0, sizeof(action));
    action.sa_handler = SIG_IGN;
    sigaction(SIGPIPE, &action, NULL);

    // Handle SIGINT signal.
    memset(&action, 0, sizeof(action));
    action.sa_handler = sigint_handler;
    sigaction(SIGINT, &action, NULL);

    return 0;
}

void sigint_handler(int signal) {
    if (signal != SIGINT)
        return;

    info("Caught SIGINT, changing state");

    if (pthread_mutex_lock(&_st_lock)) {
        warn("Mutex locking failed. Ignoring SIGINT.");
        return;
    }

    switch (_st) {
    case RUNNING:
        _st = WRAPUP;
        break;
    case WRAPUP:
        _st = END;
        break;
    case END: // don't change _st
        break;
    default:
        assert(0); // bug
    }

    while (pthread_mutex_unlock(&_st_lock))
        warn("Mutex unlocking failed. Retrying.");

    return;
}

void close_stdinout(void) {
    if (_stdinout_closed)
        return;

    fclose(stdin);
    fclose(stdout);
    _stdinout_closed = true;
    return;
}

void run_manager (
    const char * confpath,
    const char * addr,
    char * exp_prog_n_args[])
{
    fd_set rset, wset;
    int nfds, rst, cur_st, new_st;
    pktlab_time_t select_until, cur_time;
    struct timeval tv;

    assert(
        confpath != NULL && addr != NULL &&
        exp_prog_n_args != NULL);

    trace("%s(confpath:%s,addr:%s)",
        __func__, confpath, addr);

    get_st(&cur_st);
    if (cur_st != RUNNING) {
        fatal("Not in correct state before manager initialization, "
            "potential SIGINT caught beforehand");
        debug("Currently in %d", cur_st);
        exit(EXIT_FAILURE);
    }

    if (manager_init(
            confpath, addr, exp_prog_n_args) < 0) {
        fatal("Manager init failed, aborting");
        exit(EXIT_FAILURE);
    }

    if (manager_load() < 0) {
        fatal("Manager load failed, aborting");
        exit(EXIT_FAILURE);
    }

    get_st(&cur_st);
    if (cur_st != RUNNING) {
        fatal("Not in running state before manager start, aborting");
        debug("Currently in %d", cur_st);
        exit(EXIT_FAILURE);
    }

    if (manager_start() < 0) {
        fatal("Manager start failed, aborting");
        exit(EXIT_FAILURE);
    }

    // use old cur_st
    while (true) {
        FD_ZERO(&rset);
        FD_ZERO(&wset);
        select_until = UINT64_MAX; // default wait till the end of time

        nfds = manager_prepare_select(
            &rset, &wset, &select_until);
        if (nfds < 0) {
            fatal("Manager prepare select failed, aborting");
            exit(EXIT_FAILURE);
        }

        debug("Requested select until %" PRIu64, select_until);

        cur_time = pktlab_time_now();
        if (select_until < cur_time)
            select_until = cur_time;
        pktlab_time_to_timeval(select_until-cur_time, &tv);

        // if select time longer than 31 days, cap at 31
        if (tv.tv_sec >= MAX_SELECT_SEC) {
            tv.tv_sec = MAX_SELECT_SEC;
            tv.tv_usec = 0;
        }

        debug("Actual selecting period: tv->tv_sec:%u, tv->tv_usec:%u",
            tv.tv_sec, tv.tv_usec);

        rst = select(nfds, &rset, &wset, NULL, &tv);
        if (rst < 0 && errno != EINTR) {
            fatal("Main select failed, aborting");
            debug("select failed with %s", strerror(errno));
            exit(EXIT_FAILURE);
        }

        rst = manager_process_select(
            &rset, &wset, pktlab_time_now());
        if (rst < 0) {
            fatal("Manager process select failed, aborting");
            exit(EXIT_FAILURE);
        }

        if (manager_is_done())
            break;

        get_st(&new_st);
        if (new_st == cur_st)
            continue;

        // state changed!
        if (new_st == WRAPUP) {
            rst = manager_wrapup();
            if (rst < 0) {
                fatal("Manager wrap up failed, aborting");
                exit(EXIT_FAILURE);
            }
        } else if (new_st == END) {
            warn("Forcing manager to end");
            exit(EXIT_FAILURE); // kill all threads
        } else {
            assert(0); // bug
        }

        cur_st = new_st;
    }

    manager_cleanup();
    return;
}

void get_st(int * cur_st) {
    assert(cur_st != NULL);

    lock_until_success(&_st_lock);
    *cur_st = _st;
    unlock_until_success(&_st_lock);

    return;
}