// manager.c
//

#include <assert.h>
#include <inttypes.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include <netdb.h>
#include <sys/select.h>
#include <unistd.h>

#include <pktlab.h>
#include <pktctrl.h>
#include <pktconf.h>

#include "manager.h"

#include "cert.h"
#include "channel.h"
#include "config.h"
#include "console.h"
#include "ge_deputy.h"
#include "key.h"
#include "net.h"
#include "snikey.h"
#include "util.h"
#include "xdescr.h"

//
// INTERNAL CONSTANTS
//

#define DEFAULT_TARGET_BACKLOG_CONN   2
#define DEFAULT_BACKLOG_SETUP_TIMEOUT 2000000000 // 2s

#define DEFAULT_TARGET_PORT_STR "20556" // listen or proxy port
#define DEFAULT_BROKER_PORT_STR "20556"

enum manager_st {
    NOT_INIT     = 0,
    INIT         = 1,
    LOADED       = 2,
    PG_ONGOING   = 3,
    G_ONGOING    = 4,
    WRAP_UP      = 5,
    DONE         = 6
};

//
// INTERNAL TYPE DEFINITIONS
//

struct loaded_exp_exec_ll_node {
    const struct exp_exec_entry * config_entry;
    struct pktlab_certificate ** exppriv_chain;
    struct pktlab_certificate ** pub_exppriv_chain;
    struct pktlab_publickey ** trusted_eop;
    struct loaded_exp_exec_ll_node * next;
    uint_fast32_t chainlen;
    uint_fast32_t pubchainlen;
    uint_fast32_t eopcnt;
    uint_fast32_t pub_exppriv_chain_eop;
    char * sni_key;
};

//
// INTERNAL GLOBAL VARIABLES
//

static struct {
    enum manager_st st;

    const char * confpath;
    const char * addr;
    char ** exp_prog_n_args;

    struct addrinfo * broker_ai;
    struct addrinfo * addr_ai;
    struct pktlab_privatekey * controller_key;
    struct pktlab_certificate * broker_agent_cert;
    struct pktlab_certificate * proxy_agent_cert;
    struct pktlab_publickey ** broker_roots;
    struct pktlab_publickey ** proxy_roots;
    struct loaded_exp_exec_ll_node * loaded_ee_ll;
    uint_fast32_t broker_rootno;
    uint_fast32_t proxy_rootno;

    int broker_fd;
    struct pktctrl_obj * broker_ctrlobj;
    pktlab_time_t max_pub_time;

    struct ge_methods ge_module;
    struct snikey_ll_node * snikeys;
    struct pktctrl_keycert_store * greet_store;
    pktlab_time_t max_greet_time;
} _st;

//
// INTERNAL FUNCTION DECLARATIONS
//

static int publish_exp(pktlab_time_t * max_pub_time);
static int process_broker_readable(void);

static int load_loaded_ee_ll (
    const struct exp_exec_entry * exp_exec_ll,
    struct pktlab_privatekey * controller_key,
    struct loaded_exp_exec_ll_node ** loaded_ee_ll,
    uint_fast32_t * errlineno);
static void cleanup_loaded_ee_ll (
    struct loaded_exp_exec_ll_node * ll);

static int get_snikey_ll (
    struct loaded_exp_exec_ll_node * eell,
    struct snikey_ll_node ** sll);
static void free_snikey_ll(struct snikey_ll_node * sll);

static int xcert_supply_chain (
    struct pktlab_certificate ** chain,
    uint_fast32_t chainlen,
    struct pktctrl_obj * ctrlobj,
    uint16_t * errid);
static int xpub_publish (
    const char * pub_controller_addr,
    const char * sni_key,
    bool pub_period_specified,
    pktlab_time_t pub_period,
    bool pub_exp_priority_specified,
    unsigned long long pub_exp_priority,
    const struct pktconf_sep_str_ll * pub_country,
    const struct pktconf_sep_str_ll * pub_location,
    const struct exp_exec_entry * ee_entry,
    const struct pktlab_publickey * rooteop,
    struct pktctrl_obj * ctrlobj,
    pktlab_time_t * max_pub_end,
    uint16_t * errid);
static int get_server_keycert_store (
    struct loaded_exp_exec_ll_node * loaded_ee_ll,
    struct pktctrl_keycert_store ** store);

static void publish_exp_cleanup(void);

static int encode_channels (
    const struct pktconf_sep_str_ll * ll,
    const struct pktlab_publickey * rooteop,
    uint8_t ** chlistptr, uint_fast32_t * chlistlen);

static inline const char * state_str(int st);

//
// EXPORTED FUNCTION DEFINITIONS
//

int manager_init (
    const char * confpath,
    const char * addr,
    char * exp_prog_n_args[])
{
    trace("%s(_st.st:%s,confpath:%s,addr:%s)",
        __func__, state_str(_st.st),
        safe_str(confpath), safe_str(addr));

    if (_st.st != NOT_INIT || confpath == NULL ||
        addr == NULL || exp_prog_n_args == NULL)
        return -1;

    memset(&_st, 0, sizeof(_st));

    _st.confpath = confpath;
    _st.addr = addr;
    _st.exp_prog_n_args = exp_prog_n_args;
    _st.broker_fd = -1;
    _st.st = INIT;
    return 0;
}

int manager_load(void) {
    uint_fast32_t lineno;
    int rst, err, ret = -128;

    trace("%s(_st.st:%s)",
        __func__, state_str(_st.st));

    if (_st.st != INIT)
        return -1;

    // config reading
    rst = config_init();
    assert(rst == 0);

    if (config_read(_st.confpath, &lineno) < 0) {
        if (lineno == 0) {
            fatal("Failed to open config path");
            return -1;
        }

        fatal("Failed to parse config content "
            "(bad line %" PRIuFAST32 ")", lineno);
        return -1;
    } else if (!is_config_presence_correct(&err)) {
        fatal("Config option presence requirement not satisfied");
        debug("is_config_presence_correct returned err: %d", err);
        ret = -1;
        goto fail_config;
    }

    // config content parsing
    // check and parse network addresses
    if (parse_n_getaddrinfo(
            _st.addr,
            !_config.enable_proxy_mode,
            &_st.addr_ai, DEFAULT_TARGET_PORT_STR) < 0) {
        fatal("Bad commandline address string (%s)",
            safe_str(_st.addr));
        ret = -2;
        goto fail_config;
    }

    if (_config.enable_exp_pub &&
        parse_n_getaddrinfo(
            _config.broker_addr, 0,
            &_st.broker_ai, DEFAULT_BROKER_PORT_STR) < 0) {
        fatal("Bad broker address string (%s)",
            _config.broker_addr);
        ret = -3;
        goto fail_addr_ai;
    }

    // load controller key
    if (load_key_w_prompt(
            _config.controller_privkey_path,
            &_st.controller_key) < 0) {
        fatal("Failed to load controller key at %s",
            _config.controller_privkey_path);
        ret = -4;
        goto fail_broker_ai;
    }

    // load broker agent cert
    if (_config.enable_exp_pub) {
        if (load_cert(
                _config.pub_agent_cert_path,
                &_st.broker_agent_cert) < 0) {
            fatal("Failed to load agent cert for broker");
            ret = -5;
            goto fail_controller_key;
        }

        // reorder checks if there exists a cert with matching pubkey
        if (reorder_certs(
                _st.controller_key,
                &_st.broker_agent_cert, 1) < 0) {
            fatal("Agent cert (for broker) public key "
                "does not match controller key");
            ret = -6;
            goto fail_broker_agent_cert;
        }
    }

    // load broker chain roots
    if (_config.enable_exp_pub &&
        _config.broker_cert_root_path_ll != NULL) {
        if (ll_load_pubkeys(
                _config.broker_cert_root_path_ll,
                &_st.broker_roots, &_st.broker_rootno) < 0) {
            fatal("Failed to load broker identity chain roots");
            ret = -7;
            goto fail_broker_agent_cert;
        }
    }

    // load proxy agent cert
    if (_config.enable_proxy_mode &&
        _config.backlog_conn_agent_cert_path != NULL) {
        if (load_cert(
                _config.backlog_conn_agent_cert_path,
                &_st.proxy_agent_cert) < 0) {
            fatal("Failed to load agent cert for proxy");
            ret = -8;
            goto fail_broker_roots;
        }

        // reorder checks if there exists a cert with matching pubkey
        if (reorder_certs(
                _st.controller_key,
                &_st.proxy_agent_cert, 1) < 0) {
            fatal("Agent cert (for proxy) public key "
                "does not match controller key");
            ret = -9;
            goto fail_proxy_agent_cert;
        }
    }

    // load proxy chain roots
    if (_config.enable_proxy_mode &&
        _config.proxy_cert_root_path_ll != NULL) {
        if (ll_load_pubkeys(
                _config.proxy_cert_root_path_ll,
                &_st.proxy_roots, &_st.proxy_rootno) < 0) {
            fatal("Failed to load proxy agent cert roots");
            ret = -10;
            goto fail_proxy_agent_cert;
        }
    }

    // load exp_exec stuffs
    rst = load_loaded_ee_ll(
        _config.exp_exec_ll, _st.controller_key,
        &_st.loaded_ee_ll, &lineno);
    if (rst < 0) {
        fatal("Failed to load exp exec entry for line %"
            PRIuFAST32, lineno);
        debug("load_loaded_ee_ll returned %d", rst);
        ret = -11;
        goto fail_proxy_roots;
    }

    // create snikeys list
    if (_config.enable_proxy_mode &&
        get_snikey_ll(
            _st.loaded_ee_ll, &_st.snikeys) < 0) {
        fatal("Failed to get proxy register snikey list");
        ret = -12;
        goto fail_loaded_ee_ll;
    }

    info("Manager state loaded");
    _st.st = LOADED;
    return 0;

fail_loaded_ee_ll:
    cleanup_loaded_ee_ll(_st.loaded_ee_ll);
fail_proxy_roots:
    free_pubkeys(
        _st.proxy_roots,
        _st.proxy_rootno);
fail_proxy_agent_cert:
    pktlab_cleanup_cert(_st.proxy_agent_cert);
fail_broker_roots:
    free_pubkeys(
        _st.broker_roots,
        _st.broker_rootno);
fail_broker_agent_cert:
    pktlab_cleanup_cert(_st.broker_agent_cert);
fail_controller_key:
    pktlab_cleanup_privatekey(_st.controller_key);
fail_broker_ai:
    if (_st.broker_ai != NULL)
        freeaddrinfo(_st.broker_ai);
fail_addr_ai:
    if (_st.addr_ai != NULL)
        freeaddrinfo(_st.addr_ai);
fail_config:
    config_cleanup();

    // for safe manager_cleanup
    _st.addr_ai = NULL;
    _st.broker_ai = NULL;
    _st.controller_key = NULL;
    _st.broker_agent_cert = NULL;
    _st.proxy_agent_cert = NULL;
    _st.broker_roots = NULL;
    _st.proxy_roots = NULL;
    _st.loaded_ee_ll = NULL;
    _st.snikeys = NULL;
    _st.broker_rootno = 0;
    _st.proxy_rootno = 0;

    return ret;
}

int manager_start(void) {
    int rst, ret = -128;
    struct ge_passive_aux * paux;
    struct ge_active_aux * aaux;
    void * aux;

    trace("%s(_st.st:%s)",
        __func__, state_str(_st.st));

    if (_st.st != LOADED)
        return -1;

    _st.max_greet_time = UINT64_MAX; // greet indefinite if no publication
    _st.st = G_ONGOING;

    if (_config.enable_exp_pub) {
        info("Publishing experiment(s) to %s", _config.broker_addr);

        _st.st = PG_ONGOING;

        rst = publish_exp(&_st.max_pub_time);
        if (rst < 0) {
            fatal("Experiment publication failed");
            debug("publish_exp failed with %d", rst);
            ret = -2;
            goto fail;
        }

        info("Experiment publication completed");

        _st.max_greet_time = ceil_add_time(
            _st.max_pub_time, _config.pub_end_greet_period);
        if (_st.max_pub_time > UINT64_MAX-_config.pub_end_greet_period) {
            warn("Detected greeting wait end wrap around. "
                "Capping at %" PRIu64 ".", _st.max_greet_time);
        }
    }

    info("Greeting endpoints until %" PRIu64, _st.max_greet_time);

    if (get_server_keycert_store(
            _st.loaded_ee_ll, &_st.greet_store) < 0) {
        fatal("Failed to create greet endpoint keycert store");
        ret = -3;
        goto fail;
    }

    if (!_config.enable_proxy_mode_specified ||
        !_config.enable_proxy_mode) { // non-proxy (passive) mode
        info("Greeting endpoints via passive/accept method");
        _st.ge_module = get_passive_ge_methods();
        paux = safe_malloc(sizeof(*paux));
        paux->listen_ai = _st.addr_ai;
        aux = paux;
    } else {
        info("Greeting endpoints via active/proxy method");
        _st.ge_module = get_active_ge_methods();

        aaux = safe_malloc(sizeof(*aaux));
        aaux->proxy_ai = _st.addr_ai;
        aaux->agent_cert = _st.proxy_agent_cert;
        aaux->trusted_proxy_root = _st.proxy_roots;
        aaux->rootno = _st.proxy_rootno;
        // potential wrap around
        aaux->target_backlog = \
            (_config.target_backlog_conn_specified) ?
                _config.target_backlog_conn:DEFAULT_TARGET_BACKLOG_CONN;
        aaux->snikeys = _st.snikeys;
        aaux->proxy_timeout = \
            (_config.backlog_setup_timeout_specified) ?
                _config.backlog_setup_timeout:DEFAULT_BACKLOG_SETUP_TIMEOUT;;
        aux = aaux;
    }

    if (_st.ge_module.init(
            _st.exp_prog_n_args, _st.controller_key,
            _st.greet_store, aux) < 0) {
        fatal("Endpoint greeting init failed");
        free(aux);
        ret = -4;
        goto fail_store;
    }

    if (_st.ge_module.start() < 0) {
        fatal("Endpoint greeting start failed");
        ret = -5;
        goto fail_ge;
    }

    return 0;

fail_ge:
    _st.ge_module.cleanup();
fail_store:
    pktctrl_cleanup_keycert_store(_st.greet_store);
    _st.greet_store = NULL;
fail:
    _st.max_pub_time = 0;
    _st.max_greet_time = 0;
    _st.st = LOADED;
    return ret;
}

int manager_wrapup(void) {
    trace("%s(_st.st:%s)",
        __func__, state_str(_st.st));

    if (_st.st <= LOADED) {
        return -1;
    } else if (_st.st == WRAP_UP) {
        return 0; // already wrapping up
    }

    info("Wrapping up manager");
    info("Manager stop publication and wait for remaining experiments to finish");

    // need to clean stuff up and change state
    switch (_st.st) {
    case PG_ONGOING:
        publish_exp_cleanup();
        // fall through
    case G_ONGOING:
        _st.ge_module.wrapup();
        _st.st = WRAP_UP;
        break;
    default:
        assert(0); // bug
    }

    return 0;
}

int manager_prepare_select (
    fd_set * rset, fd_set * wset,
    pktlab_time_t * select_until)
{
    int nsel = 0, rst;

    trace("%s(_st.st:%s)",
        __func__, state_str(_st.st));

    if (rset == NULL || wset == NULL ||
        select_until == NULL || _st.st <= LOADED)
        return -1;

    // set fd
    switch (_st.st) {
    case PG_ONGOING:
        rst = pktctrl_prepare_select(
            &nsel, rset, wset,
            true, false, _st.broker_ctrlobj);
        assert(rst == PKTCTRL_SUCCESS);
        // fall through
    case G_ONGOING:
    case WRAP_UP:
        rst = _st.ge_module.prepare_select(
            &nsel, rset, wset, select_until);
        assert(rst == 0);
        // fall through
    case DONE: // do nothing
        break;
    default:
        assert(0); // bug
    }

    // set select time
    switch (_st.st) {
    case PG_ONGOING:
        *select_until = MIN(*select_until, _st.max_pub_time);
        break;
    case G_ONGOING:
        *select_until = MIN(*select_until, _st.max_greet_time);
        break;
    case WRAP_UP:
        // wait until the modules see fit
        // select_until not set
        break;
    case DONE:
        *select_until = 0; // no wait
        break;
    default:
        assert(0);
    }

    return nsel;
}

int manager_process_select (
    const fd_set * rset, const fd_set * wset,
    pktlab_time_t now)
{
    int rst;
    bool readable, writable;

    trace("%s(_st.st:%s,now:%" PRIu64 ")",
        __func__, state_str(_st.st), now);

    if (rset == NULL || wset == NULL ||
        _st.st <= LOADED)
        return -1;

    switch (_st.st) {
    case PG_ONGOING:
        rst = pktctrl_process_select(
            rset, wset, _st.broker_ctrlobj,
            &readable, &writable);
        assert(rst == PKTCTRL_SUCCESS);

        if (readable) {
            debug("broker_ctrlobj is ready for read");
            process_broker_readable(); // may change state
        }
        // fall through
    case G_ONGOING:
    case WRAP_UP:
        rst = _st.ge_module.process_select(rset, wset, now);
        if (rst < 0) {
            debug("Endpoint greeting process select failed with %d", rst);
            return -2;
        }
        // fall through
    case DONE: // do nothing
        break;
    default:
        assert(0); // bug
    }

    // change state
    switch (_st.st) {
    case PG_ONGOING:
        if (now >= _st.max_pub_time) {
            info("Max publication time reached (max_pub_time:%"
                PRIu64 ",cur_time:%" PRIu64 ")", _st.max_pub_time, now);
            info("Stopping publication");
            publish_exp_cleanup();
            _st.st = G_ONGOING;
        }
        break;
    case G_ONGOING:
        if (now >= _st.max_greet_time) {
            info("Max endpoint greeting time reached (max_greet_time:%"
                PRIu64 ",cur_time:%" PRIu64 ")", _st.max_greet_time, now);
            info("Stopping endpoint greeting, "
                "and waiting for remaining experiments to finish");
            rst = _st.ge_module.wrapup();
            assert(rst == 0);

            _st.st = WRAP_UP;
        }
        break;
    case WRAP_UP:
        if (_st.ge_module.is_done())
            _st.st = DONE; // manager_cleanup should be called externally
        break;
    case DONE: // do nothing
        break;
    default:
        assert(0);
    }

    return 0;
}

int manager_cleanup(void) {
    int rst;

    trace("%s(_st.st:%s)", __func__, state_str(_st.st));

    if (_st.st == NOT_INIT ||
        (_st.st >= PG_ONGOING &&
         _st.st <= WRAP_UP))
        return -1;

    info("Cleaning up manager");

    rst = _st.ge_module.cleanup();
    assert(rst == 0);

    if (_st.broker_ai != NULL)
        freeaddrinfo(_st.broker_ai);
    if (_st.addr_ai != NULL)
        freeaddrinfo(_st.addr_ai);

    pktlab_cleanup_privatekey(_st.controller_key);
    pktlab_cleanup_cert(_st.broker_agent_cert);
    pktlab_cleanup_cert(_st.proxy_agent_cert);

    cleanup_loaded_ee_ll(_st.loaded_ee_ll);
    free_snikey_ll(_st.snikeys);

    if (_st.broker_roots != NULL)
        free_pubkeys(_st.broker_roots, _st.broker_rootno);
    if (_st.proxy_roots != NULL)
        free_pubkeys(_st.proxy_roots, _st.proxy_rootno);

    publish_exp_cleanup();

    // _st.greet_store is cleaned here
    // to prevent racing condition with workers
    pktctrl_cleanup_keycert_store(_st.greet_store);

    memset(&_st, 0, sizeof(_st));
    _st.st = NOT_INIT;

    config_cleanup();
    return 0;
}

bool manager_is_done(void) {
    trace("%s(_st.st:%s)", __func__, state_str(_st.st));
    return (_st.st == DONE);
}

//
// INTERNAL FUNCTION DEFINITIONS
//

int publish_exp(pktlab_time_t * max_pub_time) {
    struct pktctrl_keycert_store * store;
    struct loaded_exp_exec_ll_node * node;
    struct pktlab_publickey * pub_exppriv_chain_root_eop;
    uint16_t errid;
    pktlab_time_t pub_time;
    int rst, ret = -128;

    assert(max_pub_time != NULL && _st.st == PG_ONGOING);

    trace("%s(...)", __func__);

    store = pktctrl_create_keycert_store();
    assert(store != NULL);

    if (pktctrl_set_client_keycert(
            store, &_st.broker_agent_cert, 1,
            _st.broker_roots, _st.broker_rootno) != PKTCTRL_SUCCESS) {
        fatal("Failed to create keycert store");
        ret = -1;
        goto fail;
    }

    if (create_connect_socket(
            _st.broker_ai, &_st.broker_fd) < 0) {
        fatal("Failed to connect to broker");
        ret = -2;
        goto fail;
    }

    if (set_keepalive(_st.broker_fd) < 0) {
        fatal("Failed to set broker socket keepalive");
        ret = -3;
        goto fail_sock;
    }

    rst = set_close_on_exec(_st.broker_fd);
    if (rst < 0) {
        fatal("Failed to set broker socket close-on-exec");
        ret = -4;
        goto fail_sock;
    }

    _st.broker_ctrlobj = pktctrl_create_obj();

    rst = pktctrl_connect(
            _st.broker_fd, PKTLAB_AUTH_AGENT, NULL,
            _st.controller_key, store, _st.broker_ctrlobj,
            NULL, NULL);
    if (rst != PKTCTRL_SUCCESS) {
        fatal("Failed to set up TLS with broker");
        debug("pktctrl_connect returned: %d", rst);
        ret = -5;
        goto fail_obj;
    }

    *max_pub_time = 0;

    for (node = _st.loaded_ee_ll; node != NULL; node = node->next) {
        pub_exppriv_chain_root_eop = NULL;

        // only provide xcert chain when trusted_eop available
        // as if not available, target channels do not include
        // key channel (ensured by config parsing process),
        // and therefore do not need xcert chain
        if (node->trusted_eop != NULL) {
            rst = xcert_supply_chain(
                node->pub_exppriv_chain,
                node->pubchainlen,
                _st.broker_ctrlobj,
                &errid);
            if (rst < 0) {
                fatal("Supply certificates to broker failed");
                debug("xcert_supply_chain failed with %d", rst);

                if (rst == -7)
                    fatal("Broker returned error: %" PRIu16, errid);
                ret = -6;
                goto fail_obj;
            }

            pub_exppriv_chain_root_eop = \
                node->trusted_eop[node->pub_exppriv_chain_eop];
        }

        rst = xpub_publish(
            _config.pub_controller_addr, node->sni_key,
            _config.pub_period_specified, _config.pub_period,
            _config.pub_exp_priority_specified, _config.pub_exp_priority,
            _config.pub_country, _config.pub_location,
            node->config_entry, pub_exppriv_chain_root_eop,
            _st.broker_ctrlobj, &pub_time, &errid);
        if (rst < 0) {
            fatal("Exp publication to broker failed");
            debug("xpub_publish failed with %d", rst);

            if (rst == -9)
                fatal("Broker returned error: %" PRIu16, errid);
            ret = -7;
            goto fail_obj;
        }

        if (pub_time > *max_pub_time)
            *max_pub_time = pub_time;
    }

    if (set_nonblock(_st.broker_fd, true) < 0) {
        fatal("Failed to set broker socket non-block");
        ret = -8;
        goto fail_obj;
    }

    // don't close broker obj and fd here if success
    pktctrl_cleanup_keycert_store(store);
    return 0;

fail_obj:
    pktctrl_close(_st.broker_ctrlobj);
fail_sock:
    close(_st.broker_fd);
fail:
    pktctrl_cleanup_keycert_store(store);

    _st.broker_fd = -1;
    _st.broker_ctrlobj = NULL;
    return ret;
}

int load_loaded_ee_ll (
    const struct exp_exec_entry * exp_exec_ll,
    struct pktlab_privatekey * controller_key,
    struct loaded_exp_exec_ll_node ** loaded_ee_ll,
    uint_fast32_t * errlineno)
{
    const struct exp_exec_entry * config_ee_entry;
    struct loaded_exp_exec_ll_node * loaded_ee_node;
    int rst, ret = -128;

    assert(
        _st.st == INIT &&
        exp_exec_ll != NULL && controller_key != NULL &&
        loaded_ee_ll != NULL && errlineno != NULL);

    trace("%s(...)", __func__);

    // load exp_exec stuffs
    config_ee_entry = _config.exp_exec_ll;
    *loaded_ee_ll = NULL;
    while (config_ee_entry != NULL) {
        loaded_ee_node = safe_malloc(sizeof(*loaded_ee_node));
        memset(loaded_ee_node, 0, sizeof(*loaded_ee_node));

        // load exppriv chain roots
        if (config_ee_entry->trusted_eop_path_ll != NULL) {
            if (ll_load_pubkeys(
                    config_ee_entry->trusted_eop_path_ll,
                    &loaded_ee_node->trusted_eop,
                    &loaded_ee_node->eopcnt) < 0) {
                debug("Failed to load trusted eop");
                *errlineno = config_ee_entry->lineno;
                ret = -1;
                free(loaded_ee_node);
                goto fail;
            }
        }

        // load exppriv chain
        assert(config_ee_entry->exppriv_chain_path_ll != NULL);
        if (ll_load_cert_chain(
                config_ee_entry->exppriv_chain_path_ll,
                &loaded_ee_node->exppriv_chain,
                &loaded_ee_node->chainlen) < 0) {
            debug("Failed to load exp privilege chain");
            *errlineno = config_ee_entry->lineno;
            ret = -2;
            if (loaded_ee_node->trusted_eop != NULL)
                free_pubkeys(
                    loaded_ee_node->trusted_eop,
                    loaded_ee_node->eopcnt);
            free(loaded_ee_node);
            goto fail;
        }

        if (reorder_certs(
                controller_key,
                loaded_ee_node->exppriv_chain,
                loaded_ee_node->chainlen) < 0) {
            debug("Failed to reorder exp privilege chain");
            *errlineno = config_ee_entry->lineno;
            ret = -3;
            if (loaded_ee_node->trusted_eop != NULL)
                free_pubkeys(
                    loaded_ee_node->trusted_eop,
                    loaded_ee_node->eopcnt);
            free_certs(
                loaded_ee_node->exppriv_chain,
                loaded_ee_node->chainlen);
            free(loaded_ee_node);
            goto fail;
        }

        // verify chain to find eop
        if (loaded_ee_node->trusted_eop != NULL) {
            rst = pktlab_verify_cert_chain(
                PKTLAB_ENDPOINT_AUTH_CONTROLLER,
                loaded_ee_node->exppriv_chain,
                loaded_ee_node->trusted_eop,
                &loaded_ee_node->pub_exppriv_chain,
                loaded_ee_node->chainlen,
                loaded_ee_node->eopcnt,
                &loaded_ee_node->pubchainlen,
                &loaded_ee_node->pub_exppriv_chain_eop);
            if (rst != PKTLAB_VERIFY_SUCCESS) {
                debug("pktlab_verify_cert_chain return %d", rst);
                *errlineno = config_ee_entry->lineno;
                ret = -4;
                free_pubkeys(
                    loaded_ee_node->trusted_eop,
                    loaded_ee_node->eopcnt);
                free_certs(
                    loaded_ee_node->exppriv_chain,
                    loaded_ee_node->chainlen);
                free(loaded_ee_node);
                goto fail;
            }
        }

        loaded_ee_node->config_entry = config_ee_entry;

        // sni_key
        if (config_ee_entry->sni_key == NULL) {
            if (pktctrl_gen_snikey(
                    &loaded_ee_node->sni_key) != PKTCTRL_SUCCESS) {
                debug("pktctrl_gen_snikey failed");
                *errlineno = config_ee_entry->lineno;
                ret = -5;
                free_pubkeys(
                    loaded_ee_node->trusted_eop,
                    loaded_ee_node->eopcnt);
                free_certs(
                    loaded_ee_node->exppriv_chain,
                    loaded_ee_node->chainlen);
                free(loaded_ee_node);
                goto fail;
            }
        } else {
            if (copy_str(
                    config_ee_entry->sni_key,
                    strlen(config_ee_entry->sni_key),
                    &loaded_ee_node->sni_key) < 0) {
                debug("copy_str failed");
                *errlineno = config_ee_entry->lineno;
                ret = -6;
                free_pubkeys(
                    loaded_ee_node->trusted_eop,
                    loaded_ee_node->eopcnt);
                free_certs(
                    loaded_ee_node->exppriv_chain,
                    loaded_ee_node->chainlen);
                free(loaded_ee_node);
                goto fail;
            }
        }

        // push new node
        loaded_ee_node->next = *loaded_ee_ll;
        *loaded_ee_ll = loaded_ee_node;

        config_ee_entry = config_ee_entry->next;
    }

    return 0;

fail:
    cleanup_loaded_ee_ll(*loaded_ee_ll);
    return ret;
}

void cleanup_loaded_ee_ll (
    struct loaded_exp_exec_ll_node * ll)
{
    struct loaded_exp_exec_ll_node * ll_node;

    trace("%s(...)", __func__);

    if (ll == NULL)
        return;

    while (ll != NULL) {
        free_certs(
            ll->exppriv_chain, ll->chainlen);
        if (ll->trusted_eop != NULL)
            free_pubkeys(
                ll->trusted_eop, ll->eopcnt);
        free(ll->sni_key);

        ll_node = ll->next;
        free(ll);
        ll = ll_node;
    }

    return;
}

int get_snikey_ll (
    struct loaded_exp_exec_ll_node * eell,
    struct snikey_ll_node ** sll)
{
    struct loaded_exp_exec_ll_node * eenode;
    struct snikey_ll_node * snode;
    size_t sklen;

    assert(sll != NULL);

    *sll = NULL;
    for (eenode = eell; eenode != NULL; eenode = eenode->next) {
        sklen = strlen(eenode->sni_key)+1;

        if (sklen > SIZE_MAX-sizeof(*snode))
            goto fail;

        snode = safe_malloc(sklen+sizeof(*snode));
        snode->snikey = (char *) (snode+1);
        snode->next = *sll;
        memcpy(snode->snikey, eenode->sni_key, sklen);

        *sll = snode;
    }

    return 0;

fail:
    free_snikey_ll(*sll);
    return -1;
}

void free_snikey_ll(struct snikey_ll_node * sll) {
    struct snikey_ll_node * node;

    while (sll != NULL) {
        node = sll->next;
        free(sll);
        sll = node;
    }

    return;
}

int xcert_supply_chain (
    struct pktlab_certificate ** chain,
    uint_fast32_t chainlen,
    struct pktctrl_obj * ctrlobj,
    uint16_t * errid)
{
    uint_fast32_t i;
    struct pktlab_message xcert_msg;
    struct pktlab_message * result_msg;
    void * der;
    uint_fast32_t derlen;
    int rst;

    assert(
        _st.st == PG_ONGOING &&
        (chain == NULL) == (chainlen == 0) &&
        ctrlobj != NULL);

    trace("%s(chainlen:%" PRIuFAST32 ")",
        __func__, chainlen);

    for (i = 0; i < chainlen; ++i) {
        if (pktlab_get_cert_der(
                chain[i], &der, &derlen) < 0) {
            debug("pktlab_get_cert_der failed");
            return -1;
        }

        if (derlen > UINT32_MAX) {
            debug("derlen too long");
            free(der);
            return -2;
        }

        xcert_msg.type = PKTLAB_XCERT_MESSAGE;
        xcert_msg.xcert.derptr = der;
        xcert_msg.xcert.derlen = derlen;

        rst = pktctrl_write_message(
            ctrlobj, &xcert_msg);
        assert(rst != 0); // blocking sockets should not have EWOULDBLOCK
        if (rst < 0) {
            debug("pktctrl_write_message failed with %d", rst);
            free(der);
            return -3;
        }

        free(der);

        rst = pktctrl_read_message(ctrlobj, &result_msg);
        assert(rst != 0);
        if (rst < 0) {
            debug("pktctrl_read_message failed with %d", rst);
            return -4;
        } else if (result_msg == NULL) {
            debug("pktctrl_read_message EOF");
            return -5;
        } else if (result_msg->type != PKTLAB_RESULT_MESSAGE) {
            debug("Bad message type (%d), expecting %d",
                result_msg->type, PKTLAB_RESULT_MESSAGE);
            free(result_msg);
            return -6;
        } else if (result_msg->result.errid != 0) {
            debug("Result message failed with %" PRIu16,
                result_msg->result.errid);
            if (errid != NULL)
                *errid = result_msg->result.errid;
            free(result_msg);
            return -7;
        }

        free(result_msg);
    }

    return 0;
}

int xpub_publish (
    const char * pub_controller_addr,
    const char * sni_key,
    bool pub_period_specified,
    pktlab_time_t pub_period,
    bool pub_exp_priority_specified,
    unsigned long long pub_exp_priority,
    const struct pktconf_sep_str_ll * pub_country,
    const struct pktconf_sep_str_ll * pub_location,
    const struct exp_exec_entry * ee_entry,
    const struct pktlab_publickey * rooteop,
    struct pktctrl_obj * ctrlobj,
    pktlab_time_t * max_pub_end,
    uint16_t * errid)
{
    uint8_t raweopkey[PKTLAB_PUBLICKEY_LEN] = {};
    uint8_t * chlistptr;
    uint_fast32_t chlistlen;
    char * xdescrptr;
    size_t xdescrlen;
    struct pktlab_message xpub_msg;
    struct pktlab_message * result_msg;
    int rst, ret = -128;

    assert(
        _st.st == PG_ONGOING &&
        pub_controller_addr != NULL &&
        sni_key != NULL && ee_entry != NULL &&
        ctrlobj != NULL && max_pub_end != NULL);

    trace("%s(pub_controller_addr:%s,sni_key:%s"
        ",pub_period_specified:%d,pub_period:%" PRIu64
        ",pub_exp_priority_specified:%d,pub_exp_priority:%llu)",
        __func__, pub_controller_addr, sni_key,
        pub_period_specified, pub_period,
        pub_exp_priority_specified, pub_exp_priority);

    chlistptr = NULL;
    xdescrptr = NULL;
    if (rooteop != NULL) {
        if (pktlab_get_publickey_bytes(
                rooteop, raweopkey, sizeof(raweopkey)) < 0) {
            fatal("Cannot get public key bytes");
            return -1;
        }
    }

    rst = encode_channels(
        ee_entry->pub_channel_ll, rooteop,
        &chlistptr, &chlistlen);
    if (rst < 0) {
        fatal("Cannot generate chlist");
        debug("encode_channels returned %d", rst);
        return -2;
    }

    if (chlistlen > UINT32_MAX) {
        debug("Overflow");
        ret = -3;
        goto fail_chlist;
    }

    rst = create_xdescr(
        pub_controller_addr, sni_key,
        pub_period_specified, pub_period,
        pub_exp_priority_specified, pub_exp_priority,
        pub_country, pub_location, ee_entry,
        &xdescrptr, max_pub_end);
    if (rst < 0) {
        fatal("Cannot generate xdescr");
        debug("create_xdescr returned %d", rst);
        ret = -4;
        goto fail_chlist;
    }

    xdescrlen = strlen(xdescrptr);
    if (xdescrlen > UINT32_MAX) {
        debug("Overflow");
        ret = -3;
        goto fail_xdescr;
    }

    xpub_msg.type = PKTLAB_XPUB_MESSAGE;
    xpub_msg.xpub.chlistptr = chlistptr;
    xpub_msg.xpub.chlistlen = chlistlen;
    xpub_msg.xpub.eopkey = raweopkey;
    xpub_msg.xpub.xdescrptr = xdescrptr;
    xpub_msg.xpub.xdescrlen = xdescrlen;

    rst = pktctrl_write_message(ctrlobj, &xpub_msg);
    assert(rst != 0);
    if (rst < 0) {
        debug("pktctrl_write_message failed with %d", rst);
        ret = -5;
        goto fail_xdescr;
    }

    rst = pktctrl_read_message(ctrlobj, &result_msg);
    assert(rst != 0);
    if (rst < 0) {
        debug("pktctrl_read_message failed with %d", rst);
        ret = -6;
        goto fail_xdescr;
    } else if (result_msg == NULL) {
        debug("pktctrl_read_message EOF");
        ret = -7;
        goto fail_xdescr;
    } else if (result_msg->type != PKTLAB_RESULT_MESSAGE) {
        debug("Bad message type (%d), expecting %d",
            result_msg->type, PKTLAB_RESULT_MESSAGE);
        ret = -8;
        goto fail_result;
    } else if (result_msg->result.errid != 0) {
        debug("Result message failed with %" PRIu16,
            result_msg->result.errid);
        if (errid != NULL)
            *errid = result_msg->result.errid;
        ret = -9;
        goto fail_result;
    }

    ret = 0;

fail_result:
    free(result_msg);
fail_xdescr:
    free(xdescrptr);
fail_chlist:
    free(chlistptr);
    return ret;
}

int get_server_keycert_store (
    struct loaded_exp_exec_ll_node * loaded_ee_ll,
    struct pktctrl_keycert_store ** store)
{
    struct loaded_exp_exec_ll_node * node;
    int rst;

    assert(loaded_ee_ll != NULL && store != NULL);

    trace("%s(...)", __func__);

    *store = pktctrl_create_keycert_store();
    assert(*store != NULL);

    for (node = loaded_ee_ll; node != NULL; node = node->next) {
        assert(
            node->sni_key != NULL &&
            node->exppriv_chain != NULL && node->chainlen != 0 &&
            (node->trusted_eop == NULL) == (node->eopcnt == 0));

        rst = pktctrl_set_server_keycert(
            *store, node->sni_key,
            node->exppriv_chain, node->chainlen,
            node->trusted_eop, node->eopcnt);
        if (rst != PKTCTRL_SUCCESS) {
            debug("pktctrl_set_server_keycert failed with %d", rst);
            pktctrl_cleanup_keycert_store(*store);
            return -1;
        }

        if (!node->config_entry->is_default)
            continue;

        rst = pktctrl_set_server_keycert_default(
            *store, node->sni_key);
        assert(rst == PKTCTRL_SUCCESS);
    }

    return 0;
}

int process_broker_readable(void) {
    struct pktlab_message * msg;
    int rst, ret = -128;

    // this should only be called on a non-blocking socket

    assert(_st.st == PG_ONGOING);

    trace("%s()", __func__);

    rst = pktctrl_read_message(_st.broker_ctrlobj, &msg);
    if (rst == 0) {
        debug("pktctrl_read_message got EWOULDBLOCK");
        return -1;
    } else if (rst < 0) {
        debug("pktctrl_read_message failed with %d", rst);
        ret = -2;
        goto fail;
    } else if (msg != NULL) {
        warn("Unexpected message from broker with type %d, ignoring",
            msg->type);
        free(msg);
        return -3; // does not clean up
    }

    ret = 0;
    info("broker EOF, cleaning up");

fail:
    publish_exp_cleanup();
    _st.st = G_ONGOING;
    return ret;
}

void publish_exp_cleanup(void) {
    assert(
        _st.st != NOT_INIT &&
        (_st.broker_ctrlobj == NULL) ==
        (_st.broker_fd < 0));

    trace("%s()", __func__);

    if (_st.broker_ctrlobj == NULL)
        return;

    pktctrl_close(_st.broker_ctrlobj);
    close(_st.broker_fd);

    _st.broker_ctrlobj = NULL;
    _st.max_pub_time = 0;
    _st.broker_fd = -1;
}

int encode_channels (
    const struct pktconf_sep_str_ll * ll,
    const struct pktlab_publickey * rooteop,
    uint8_t ** chlistptr, uint_fast32_t * chlistlen)
{
    struct pktlab_channel * channels;
    int rst, ret = -128;

    assert(chlistptr != NULL && chlistlen != NULL);

    trace("%s(...)", __func__);

    rst = get_pktlab_channels(ll, rooteop, &channels);
    if (rst < 0) {
        debug("get_pktlab_channels failed with %d", rst);
        assert(rst != -1);
        return -1;
    }

    assert(channels != NULL);
    *chlistptr = NULL;
    *chlistlen = 0;

    rst = pktlab_encode_chlist(
        channels, (void **) chlistptr, chlistlen);
    if (rst < 0) {
        debug("pktlab_encode_chlist failed with %d", rst);
        ret = -2;
        goto fail_channels;
    }

    ret = 0;

fail_channels:
    free_pktlab_channels(channels);
    return ret;
}

const char * state_str(int st) {
    static const char * sstr[] = {
        [NOT_INIT]     = "NOT_INIT",
        [INIT]         = "INIT",
        [LOADED]       = "LOADED",
        [PG_ONGOING]   = "PG_ONGOING",
        [G_ONGOING]    = "G_ONGOING",
        [WRAP_UP]      = "WRAP_UP",
        [DONE]         = "DONE"
    };

    assert(st >= 0 && st < sizeof(sstr)/sizeof(*sstr));
    return safe_str(sstr[st]);
}