// net.c
//

#include <assert.h>
#include <errno.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <fcntl.h>
#include <netdb.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include "console.h"
#include "net.h"
#include "util.h"

//
// EXPORTED FUNCTION DEFINITIONS
//

int parse_host_port (
    const char * hostport,
    char ** host, char ** port)
{
    size_t hplen, hlen, plen, i;
    const char * ptr;

    assert(hostport != NULL && host != NULL && port != NULL);

    hplen = strlen(hostport);
    if (hplen == 0)
        return -1;

    // ipv6 with port case
    if (hostport[0] == '[') {
        hlen = 0;
        for (i = 1; i < hplen; ++i) {
            if (hostport[i] == ']')
                break;
            ++hlen;
        }

        if (hostport[i] != ']' ||
            hostport[1+hlen+1] != ':')
            return -1; // ']' not found, or port not at expected location

        plen = hplen - (1+hlen+1+1); // overflow may happen
        if (hlen == 0 || plen == 0 ||
            hlen >= SIZE_MAX || plen >= SIZE_MAX)
            return -1;

        copy_str(hostport+1, hlen, host);
        copy_str(hostport+1+hlen+1+1, plen, port);
        return 0;
    }

    // ipv6 case
    if (hplen >= 5 && hostport[4] == ':') {
        copy_str(hostport, hplen, host);
        *port = NULL;
        return 0;
    }

    ptr = strchr_ext(hostport, ':', &hlen);
    if (ptr == NULL) { // ipv4/domain with port case
        copy_str(hostport, hplen, host);
        *port = NULL;
        return 0;
    }

    // ipv4/domain case
    plen = hplen-hlen-1;
    copy_str(hostport, hlen, host);
    copy_str(hostport+hlen+1, plen, port);
    return 0;
}

int parse_n_getaddrinfo (
    const char * hostport, int for_bind,
    struct addrinfo ** ai, const char * default_port_str)
{
    char * host;
    char * port;
    struct addrinfo hints;
    struct addrinfo * res;
    int rst;

    assert(hostport != NULL);

    if (parse_host_port(hostport, &host, &port))
        return -1;

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_flags = for_bind ? AI_PASSIVE:0;

    hints.ai_flags |= AI_NUMERICSERV|AI_ADDRCONFIG;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;

    rst = getaddrinfo(
        host, (port == NULL) ?
        default_port_str:port, &hints, &res);
    if (rst != 0) {
        debug("Address %s (host:%s,port:%s) getaddrinfo failed with %d",
            hostport, safe_str(host), safe_str(port), rst);
        free(host);
        free(port);
        return -1;
    }
    assert(res != NULL);

    if (res->ai_next != NULL) {
        freeaddrinfo(res->ai_next);
        res->ai_next = NULL;
    }
    if (ai != NULL) {
        *ai = res;
    } else {
        freeaddrinfo(res);
    }

    free(host);
    free(port);
    return 0;
}

int create_connect_socket (
    struct addrinfo * ai, int * sock)
{
    assert(ai != NULL && sock != NULL);

    *sock = socket(
        ai->ai_family,
        ai->ai_socktype,
        ai->ai_protocol);
    if (*sock < 0) {
        debug("socket returned %s", strerror(errno));
        return -1;
    }

    if (connect(
            *sock, ai->ai_addr,
            ai->ai_addrlen) < 0) {
        debug("connect returned %s", strerror(errno));
        close(*sock);
        return -1;
    }

    return 0;
}

int setup_nonblock_connect_socket (
    struct addrinfo * ai, int * sock, int * err)
{
    int ret = -128, rst;

    assert(ai != NULL && sock != NULL);

    *sock = socket(
        ai->ai_family,
        ai->ai_socktype,
        ai->ai_protocol);
    if (*sock < 0) {
        debug("socket returned %s", strerror(errno));
        return -1;
    }

    if (set_nonblock(*sock, true) < 0) {
        debug("set_nonblock(%d) failed", *sock);
        ret = -2;
        goto fail_sock;
    }

    rst = connect(
        *sock, ai->ai_addr, ai->ai_addrlen);
    if (rst == 0) {
        return 1; // connect done surprisingly fast
    } else if (rst < 0 && errno != EINPROGRESS) {
        debug("connect(%d) failed with: %s",
            *sock, strerror(errno));
        if (err != NULL)
            *err = errno;
        ret = -3;
        goto fail_sock;
    }

    return 0;

fail_sock:
    close(*sock);
    return ret;
}

int check_nonblock_connect_socket(int sock, int * err) {
    int rst;
    socklen_t optlen;

    assert(sock >= 0);

    optlen = sizeof(rst);
    if (getsockopt(
            sock, SOL_SOCKET, SO_ERROR,
            &rst, &optlen) < 0) { // timeout
        debug("getsockopt(%d,SO_ERROR) failed with: %s",
            sock, strerror(errno));
        return -1;
    }

    if (rst != 0) {
        debug("connect failed with: %s", strerror(rst));
        if (err != NULL)
            *err = rst;
        return -2;
    }

    return 0; // connect completed
}

int create_welcome_socket (
    struct addrinfo * ai, int * sock)
{
    assert(ai != NULL && sock != NULL);

    *sock = socket(
        ai->ai_family,
        ai->ai_socktype,
        ai->ai_protocol);
    if (*sock < 0) {
        debug("socket returned %s", strerror(errno));
        return -1;
    }

    if (bind(*sock, ai->ai_addr, ai->ai_addrlen) < 0) {
        debug("bind returned %s", strerror(errno));
        close(*sock);
        return -1;
    }

    if (listen(*sock, 0) < 0) {
        debug("listen returned %s", strerror(errno));
        close(*sock);
        return -1;
    }

    return 0;
}

int accept_new_socket (
    int welcome_sock, int * new_sock,
    struct sockaddr_storage * sa, int * err)
{
    socklen_t salen = sizeof(*sa);

    assert(
        welcome_sock >= 0 && new_sock != NULL &&
        sa != NULL && err != NULL);

    *new_sock = accept(
        welcome_sock, (struct sockaddr *) sa, (socklen_t *) &salen);
    if (*new_sock < 0) {
        debug("accept failed with %s", strerror(errno));
        *err = errno;
        return -1;
    }

    return 0;
}

int get_stream_socketpair(int fds[2], int * err) {
    assert(fds != NULL && err != NULL);

    if (socketpair(AF_UNIX, SOCK_STREAM, 0, fds) < 0) {
        debug("socketpair failed with %s", strerror(errno));
        *err = errno;
        return -1;
    }

    return 0;
}

int set_keepalive(int sktfd) {
    int optval;

    assert(sktfd >= 0);

    optval = 1;
    if (setsockopt(
            sktfd, SOL_SOCKET, SO_KEEPALIVE,
            (void *) &optval, (socklen_t) sizeof(optval)) < 0) {
        debug("setsockopt returned %s", strerror(errno));
        return -1;
    }

    return 0;

}

int set_nonblock(int sktfd, bool set) {
    int old_flag, new_flag;

    assert(sktfd >= 0);

    trace("%s(sktfd:%d,set:%d)", __func__, sktfd, (int) set);

    old_flag = fcntl(sktfd, F_GETFL);
    if (old_flag < 0) {
        debug("fcntl(F_GETFL) returned %s", strerror(errno));
        return -1;
    }

    new_flag = set ? (old_flag|O_NONBLOCK):(old_flag&~O_NONBLOCK);

    if (fcntl(sktfd, F_SETFL, new_flag) < 0) {
        debug("fcntl(F_SETFL) returned %s", strerror(errno));
        return -1;
    }

    return 0;
}

int print_info_addr(const char * prestr, struct sockaddr_storage * sa) {
    // INET6_ADDRSTRLEN larger than INET_ADDRSTRLEN
    char addrstr[INET6_ADDRSTRLEN];
    int af;
    void * src;

    assert(prestr != NULL && sa != NULL);

    af = sa->ss_family;

    switch (sa->ss_family) {
    case AF_INET:
        src = &(((struct sockaddr_in *) sa)->sin_addr.s_addr);
        break;
    case AF_INET6:
        src = ((struct sockaddr_in6 *) sa)->sin6_addr.s6_addr;
        break;
    default:
        return -1; // unknown addr type
    }

    if (inet_ntop(af, src, addrstr, sizeof(addrstr)) == 0) {
        debug("inet_ntop failed to convert addr");
        return -2;
    }

    info("%s%s", prestr, addrstr);
    return 0;
}