// util.c
//

#include "util.h"

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <inttypes.h>
#include <limits.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <fcntl.h>
#include <pthread.h>
#include <sys/stat.h>
#include <termios.h>

#include <pktconf.h>
#include <pktctrl.h>

#include "console.h"

//
// INTERNAL CONSTANTS
//

//
// EXPORTED FUNCTION DEFINITIONS
//

void * pktlabme_safe_malloc_(size_t size) {
    void * ptr;

    ptr = malloc(size);

    if (ptr == NULL) {
        perror(NULL);
        exit(EXIT_FAILURE);
        return NULL;
    } else
        return ptr;
}

void * pktlabme_safe_calloc_(size_t nelts, size_t eltsz) {
    void * ptr;

    ptr = calloc(nelts, eltsz);

    if (ptr == NULL) {
        perror(NULL);
        exit(EXIT_FAILURE);
        return NULL;
    } else
        return ptr;
}

void * pktlabme_safe_realloc_(void * ptr, size_t size) {

    ptr = realloc(ptr, size);

    if (ptr == NULL && size > 0) {
        perror(NULL);
        exit(EXIT_FAILURE);
        return NULL;
    } else
        return ptr;
}

const char * strchr_ext(const char * s, int c, size_t * l) {
    const char * ptr;

    if (s == NULL)
        return NULL;

    if (l != NULL)
        *l = 0;

    for (ptr = s; *ptr != '\0'; ++ptr) {
        if (*ptr == c)
            return ptr;

        if (l != NULL) {
            ++*l;

            if (*l == SIZE_MAX)
                break; // *l will overflow
        }
    }

    return NULL;
}

int copy_str(const char * s, size_t l, char ** copied_str) {
    assert(s != NULL && copied_str != NULL);

    if (l >= SIZE_MAX)
        return -1;

    *copied_str = safe_malloc(l+1);
    memcpy(*copied_str, s, l);
    (*copied_str)[l] = '\0';

    return 0;
}

int read_whole_file (
    const char * path,
    uint8_t ** content, size_t * contentlen)
{
    FILE * fp;
    uint8_t * ptr;
    size_t total, incre, rst;

    assert(
        path != NULL && content != NULL &&
        contentlen != NULL);

    trace("%s(path:%s)", __func__, path);

    fp = fopen(path, "r");
    if (fp == NULL)
        return -1;

    total = incre = 4096;
    *contentlen = 0;
    *content = ptr = safe_malloc(total);
    while (true) {
        assert(total >= *contentlen);
        rst = fread(ptr, 1, total-*contentlen, fp);
        if (ferror(fp)) {
            debug("fread failed with: %s", strerror(errno));
            goto fail;
        }

        *contentlen += rst;
        ptr += rst;

        if (feof(fp))
            break;

        if (total-*contentlen < incre) {
            if (total > SIZE_MAX-incre) {
                debug("Overflow");
                goto fail;
            }

            total += incre;
            safe_realloc(ptr, total);
        }
    }

    fclose(fp);
    return 0;

fail:
    free(ptr);
    fclose(fp);
    return -1;
}

int posix_getpass (
    const char * prompt, char ** passphrase)
{
    FILE * fp;
    int ret = -1;
    struct termios oflags, nflags;
    const size_t max_pplen = 512; // smaller then INT_MAX

    assert(passphrase != NULL);

    fp = fopen("/dev/tty", "r+");
    if (fp == NULL) {
        debug("fopen failed with %s", strerror(errno));
        return -1;
    }

    if (tcgetattr(fileno(fp), &oflags) < 0){
        debug("tcgetattr failed with %s", strerror(errno));
        goto fail;
    }

    nflags = oflags;
    nflags.c_lflag &= ~ECHO;
    nflags.c_lflag |= ECHONL;

    if (tcsetattr(fileno(fp), TCSANOW, &nflags) < 0) {
        debug("tcsetattr failed with %s", strerror(errno));
        goto fail;
    }

    if (prompt != NULL && !fputs(prompt, fp)) {
        debug("fputs failed with %s", strerror(errno));
        goto fail;
    }

    *passphrase = safe_malloc(max_pplen);
    if (fgets(*passphrase, max_pplen, fp) == NULL) {
        debug("fgets failed with %s", strerror(errno));
        free(*passphrase);
        goto fail;
    }

    (*passphrase)[strcspn(*passphrase, "\n")] = '\0';

    if (tcsetattr(fileno(fp), TCSANOW, &oflags) < 0) {
        debug("tcsetattr failed with %s", strerror(errno));
        free(*passphrase);
        goto fail;
    }

    ret = 0;

fail:
    fclose(fp);
    return ret;
}

bool is_valid_pktlab_time(const char * pt, pktlab_time_t * rst) {
    unsigned long long ull;

    if (!is_valid_ull(pt, &ull))
        return false;

    if (ull > UINT64_MAX)
        return false;

    if (rst != NULL)
        *rst = ull;

    return true;
}

bool is_valid_ull (
    const char * ull, unsigned long long * rst)
{
    size_t len;
    unsigned long long tmp_ull;
    char * endptr;

    assert(ull != NULL);

    len = strlen(ull);

    errno = 0;
    tmp_ull = strtoull(ull, &endptr, 10);
    if (errno != 0 || endptr != ull+len) // don't allow "final string"
        return false;

    if (rst != NULL)
        *rst = tmp_ull;

    return true;
}

bool is_valid_filepath(const char * path) {
    struct stat st;

    assert(path != NULL);

    if (stat(path, &st) < 0) {
        debug("stat on %s failed with error %s",
            path, strerror(errno));
        return false;
    }

    // accept if is regular file or symbolic link
    return (S_ISREG(st.st_mode) || S_ISLNK(st.st_mode));
}

bool is_valid_iso31661_alpha3(const char * country) {
    size_t len, i;

    assert(country != NULL);

    len = strlen(country);
    if (len != 3)
        return false;

    for (i = 0; i < len; ++i)
        if (!isupper(country[i]))
            return false;

    return true;
}

bool is_valid_maidenhead_w_wc(const char * locator) {
    size_t len, i;

    assert(locator != NULL);

    len = strlen(locator);
    if (len == 0)
        return false;

    for (i = 0; i < len; ++i) {
        switch (locator[i]) {
        case '*':
        case '?':
            break;
        default:
            if (!isupper(locator[i]) &&
                !isdigit(locator[i]))
                return false;
        }
    }

    return true;
}

bool is_valid_hexstr(const char * hexstr) {
    size_t len, i;

    assert(hexstr != NULL);

    len = strlen(hexstr);
    if (len == 0 || len % 2 != 0)
        return false;

    // 0x at start not allowed
    for (i = 0; i < len; ++i) {
        if (!isxdigit(hexstr[i]))
            return false;
    }

    return true;
}

int overflow_safe_add_uf32 (
    uint_fast32_t u1, uint_fast32_t u2,
    uint_fast32_t * rst)
{
    if (u1 > UINT_FAST32_MAX-u2)
        return -1;

    *rst = u1+u2;
    return 0;
}

int get_hex_val(char c) {
    if (!isxdigit(c))
        return -1;

    if (isdigit(c))
        return c-'0';

    return tolower(c)-'a'+10;
}

int hexstr2bytes (
    const char * hexstr, size_t len,
    uint8_t ** bytes, size_t * blen)
{
    size_t i;
    int u, l;

    assert(
        hexstr != NULL && len != 0 &&
        len % 2 == 0 && bytes != NULL);

    *blen = len/2;
    *bytes = safe_malloc(*blen);

    for (i = 0; i < len; i += 2) {
        u = get_hex_val(hexstr[i]);
        l = get_hex_val(hexstr[i+1]);
        assert(u >= 0 && l >= 0 && u < 16 && l < 16);

        (*bytes)[i/2] = 16*u+l;
    }

    return 0;
}

pktlab_time_t ceil_add_time (
    pktlab_time_t i, pktlab_time_t j)
{
    if (i > UINT64_MAX-j)
        return UINT64_MAX;
    return i+j;
}

int set_close_on_exec(int fd) {
    int flags;

    assert(fd >= 0);

    flags = fcntl(fd, F_GETFD);
    if (flags < 0) {
        debug("fcntl(%d, F_GETFD) failed with %s",
            fd, strerror(errno));
        return -1;
    }

    if (fcntl(fd, F_SETFD, flags|FD_CLOEXEC) == -1) {
        debug("fcntl(%d, F_SETFD, %d) failed with %s",
            fd, flags|FD_CLOEXEC, strerror(errno));
        return -2;
    }

    return 0;
}

void lock_until_success(pthread_mutex_t * mutex) {
    int rst;

    assert(mutex != NULL);

    while (true) {
        rst = pthread_mutex_lock(mutex);
        assert(rst != EINVAL && rst != EBUSY);

        if (rst == 0)
            break;
    }

    return;
}

void unlock_until_success(pthread_mutex_t * mutex) {
    int rst;

    assert(mutex != NULL);

    while (true) {
        rst = pthread_mutex_unlock(mutex);
        assert(rst != EINVAL && rst != EPERM);

        if (rst == 0)
            break;
    }

    return;
}

void set_fd_set(int * nsel, fd_set * set, int fd) {
    assert(nsel != NULL && set !=  NULL && fd < INT_MAX);

    trace("%s(fd:%d)", __func__, fd);

    FD_SET(fd, set);
    *nsel = MAX(*nsel, fd+1);
    return;
}


int set_sigmask_block_sigint(sigset_t * oset) {
    sigset_t set;
    int rst;

    assert(oset != NULL);

    if (sigemptyset(&set) < 0) {
        debug("sigemptyset failed with %s", strerror(errno));
        return -1;
    }

    if (sigaddset(&set, SIGINT) < 0) {
        debug("sigaddset failed with %s", strerror(errno));
        return -2;
    }

    rst = pthread_sigmask(SIG_BLOCK, &set, oset);
    if (rst != 0) {
        debug("pthread_sigmask failed with %s", strerror(rst));
        return -3;
    }

    return 0;
}

int restore_sigmask(sigset_t * oset) {
    int rst;

    assert(oset != NULL);

    rst = pthread_sigmask(SIG_SETMASK, oset, NULL);
    if (rst != 0) {
        debug("pthread_sigmask failed with %s", strerror(rst));
        return -1;
    }

    return 0;
}