// worker.c
//

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <limits.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <sys/select.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include <pktlab.h>
#include <pktctrl.h>

#include "console.h"
#include "net.h"
#include "util.h"
#include "worker.h"

//
// EXPORTED FUNCTION DEFINITIONS
//

void * worker_run(void * arg) {
    struct worker_arg * wa = arg;
    int endpoint_fd, manager_fd, rst, fds[2], err, nsel;
    uint_fast32_t worker_id;
    char ** exp_prog_n_args;
    char * exp_prog;
    char * ptr;
    struct pktlab_privatekey * k;
    struct pktctrl_keycert_store * store;
    struct pktctrl_obj * endpointobj;
    struct pktctrl_obj * expprogobj;
    pid_t pid;
    fd_set rset, wset;
    struct pktlab_message * endpoint2expprog;
    struct pktlab_message * expprog2endpoint;
    bool read_from_endpoint, write_to_endpoint;
    bool read_from_expprog, write_to_expprog;
    bool endpoint_readable, endpoint_writable;
    bool expprog_readable, expprog_writable;

    assert(arg != NULL);

    trace("%s(arg:{endpoint_fd:%d,manager_fd:%d,worker_id:%"
        PRIuFAST32 ",exp_prog:%s})", __func__,
        wa->endpoint_fd, wa->manager_fd,
        wa->worker_id, safe_str(wa->exp_prog_n_args[0]));

    // extract arguments
    endpoint_fd = wa->endpoint_fd;
    manager_fd = wa->manager_fd;
    worker_id = wa->worker_id;
    exp_prog_n_args = wa->exp_prog_n_args;
    exp_prog = exp_prog_n_args[0];
    k = wa->k;
    store = wa->store;
    free(arg);
    arg = NULL;
    assert(
        endpoint_fd >= 0 && manager_fd >= 0 &&
        exp_prog_n_args != NULL && exp_prog != NULL &&
        k != NULL && store != NULL);

    // set up pktctrl
    endpointobj = pktctrl_create_obj();
    assert(endpointobj != NULL);

    rst = pktctrl_accept(
            endpoint_fd, PKTLAB_CONTROLLER_AUTH_ENDPOINT,
            k, store, endpointobj, NULL, NULL);
    if (rst != PKTCTRL_SUCCESS && rst != PKTCTRL_IN_PROGRESS) {
        warn("Worker %" PRIuFAST32
            " failed to establish TLS connection with endpoint,"
            " skipping endpoint", worker_id);
        debug("pktctrl_accept failed with %d", rst);
        goto fail;
    }

    if (rst != PKTCTRL_SUCCESS) { // in progress
        while (true) {
            FD_ZERO(&rset);
            FD_ZERO(&wset);

            FD_SET(manager_fd, &rset); // for checking eof
            nsel = manager_fd+1;

            if (pktctrl_prepare_select(
                    &nsel, &rset, &wset,
                    true, true, endpointobj) != PKTCTRL_SUCCESS) {
                warn("Worker %" PRIuFAST32
                    " failed to select pktctrl_obj for endpoint,"
                    " skipping this endpoint", worker_id);
                goto fail;
            }

            rst = select(nsel, &rset, &wset, NULL, NULL); // block indefinitely
            if (rst < 0) {
                if (errno == EINTR)
                    continue; // re-select again

                warn("Worker %" PRIuFAST32
                    " select failed (%s), skipping this endpoint",
                    worker_id, strerror(err));
                goto fail;
            }

            if (pktctrl_process_select(
                    &rset, &wset, endpointobj,
                    &endpoint_readable,
                    &endpoint_writable) != PKTCTRL_SUCCESS) {
                warn("Worker %" PRIuFAST32
                    " failed to select pktctrl_obj for endpoint,"
                    " skipping this endpoint", worker_id);
                goto fail;
            }

            if (endpoint_readable || endpoint_writable) {
                rst = pktctrl_accept(
                        endpoint_fd, PKTLAB_CONTROLLER_AUTH_ENDPOINT,
                        NULL, NULL, endpointobj, NULL, NULL);
                if (rst != PKTCTRL_SUCCESS && rst != PKTCTRL_IN_PROGRESS) {
                    warn("Worker %" PRIuFAST32
                        " failed to establish TLS connection with endpoint,"
                        " skipping endpoint", worker_id);
                    debug("pktctrl_accept failed with %d", rst);
                    goto fail;
                }

                if (rst == PKTCTRL_SUCCESS)
                    break;
            }

            if (FD_ISSET(manager_fd, &rset)) { // eof
                info("Worker %" PRIuFAST32
                    " detect manager EOF, ending this session prematurely",
                    worker_id);
                goto fail;
            }
        }
    }

    // signal pktctrl setup success
    rst = send_worker_msg(
        manager_fd, WORKER_TLS_COMPLETED);
    if (rst != 1) {
        debug("send_worker_msg returned %d", rst);
        assert(rst <= 0);

        if (rst <= 0) {
            warn("Worker %" PRIuFAST32
                " failed to notify manager on TLS connection completed,"
                " skipping endpoint", worker_id);
            debug("send_worker_msg err: %s",
                (rst == 0) ? "n/a":strerror(errno));
        }

        goto fail;
    }

    // set up exp prog socket pair
    // fds[0] is worker end
    // fds[1] is exp prog end
    if (get_stream_socketpair(fds, &err) < 0) {
        warn("Worker %" PRIuFAST32
            " failed to create worker-expprog socket pair (%s),"
            " skipping this endpoint", worker_id, strerror(err));
        goto fail;
    }

    pid = 0; // for skipping child waiting in case of early cleanup
    expprogobj = pktctrl_create_obj();
    assert(expprogobj != NULL);

    if (pktctrl_raw_session(
            fds[0], expprogobj) != PKTCTRL_SUCCESS) {
        warn("Worker %" PRIuFAST32
            " failed create pktctrl_obj for exp program (%s),"
            " skipping this endpoint", worker_id, strerror(err));
        goto fail_expprogobj;
    }

    pid = fork();
    if (pid == 0) { // child
        // move fd and exec
        close(fds[0]);

        if (dup2(fds[1], 0) < 0) {
            warn("Worker %" PRIuFAST32
                " failed to move worker descriptor to 0 (%s),"
                " skipping this endpoint", worker_id, strerror(err));
            exit(EXIT_FAILURE);
        }

        // modification to exp_prog_n_args here is fine as it is a new copy
        exp_prog = exp_prog_n_args[0];
        ptr = strrchr(exp_prog_n_args[0], '/');
        if (ptr != NULL)
            exp_prog_n_args[0] = ptr+1; // remove string prior to '/'

        // note it is possible for new exp_prog_n_args[0] to be 0 len
        // bad bin though, will fail next step

        if (execv(exp_prog, exp_prog_n_args) < 0) {
            warn("Worker %" PRIuFAST32
                " cannot run %s (%s), skipping this endpoint",
                worker_id, exp_prog, strerror(err));
            exit(EXIT_FAILURE);
        }
    }

    // parent forward traffic, and clean up when EOF
    close(fds[1]);

    info("Worker %" PRIuFAST32 " setup complete, start message forwarding",
        worker_id);

    // we only do single message buffer for each direction
    // should be NULL if no msg buffered or msg ptr
    endpoint2expprog = NULL;
    expprog2endpoint = NULL;

    while (true) {
        FD_ZERO(&rset);
        FD_ZERO(&wset);
        nsel = 0;

        if (endpoint2expprog == NULL) {
            read_from_endpoint = true;
            write_to_expprog = false;
        } else {
            read_from_endpoint = false;
            write_to_expprog = true;
        }

        if (expprog2endpoint == NULL) {
            read_from_expprog = true;
            write_to_endpoint = false;
        } else {
            read_from_expprog = false;
            write_to_endpoint = true;
        }

        if (pktctrl_write_unsent(endpointobj) > 0)
            write_to_endpoint = true;

        if (pktctrl_write_unsent(expprogobj) > 0)
            write_to_expprog = true;

        if (pktctrl_prepare_select(
                &nsel, &rset, &wset,
                read_from_endpoint,
                write_to_endpoint,
                endpointobj) != PKTCTRL_SUCCESS) {
            warn("Worker %" PRIuFAST32
                " failed to select pktctrl_obj for endpoint,"
                " skipping this endpoint", worker_id);
            goto fail_msgbuf;
        }

        if (pktctrl_prepare_select(
                &nsel, &rset, &wset,
                read_from_expprog,
                write_to_expprog,
                expprogobj) != PKTCTRL_SUCCESS) {
            warn("Worker %" PRIuFAST32
                " failed to select pktctrl_obj for exp program,"
                " skipping this endpoint", worker_id);
            goto fail_msgbuf;
        }

        FD_SET(manager_fd, &rset); // for check eof only

        rst = select(nsel, &rset, &wset, NULL, NULL); // block indefinitely
        if (rst < 0) {
            if (errno == EINTR)
                continue; // re-select again

            warn("Worker %" PRIuFAST32
                " select failed (%s), skipping this endpoint",
                worker_id, strerror(err));
            goto fail_msgbuf;
        }

        if (pktctrl_process_select(
                &rset, &wset, endpointobj,
                &endpoint_readable,
                &endpoint_writable) != PKTCTRL_SUCCESS) {
            warn("Worker %" PRIuFAST32
                " failed to select pktctrl_obj for endpoint,"
                " skipping this endpoint", worker_id);
            goto fail_msgbuf;
        }

        if (pktctrl_process_select(
                &rset, &wset, expprogobj,
                &expprog_readable,
                &expprog_writable) != PKTCTRL_SUCCESS) {
            warn("Worker %" PRIuFAST32
                " failed to select pktctrl_obj for exp program,"
                " skipping this endpoint", worker_id);
            goto fail_msgbuf;
        }

        if (endpoint_readable) {
            assert(endpoint2expprog == NULL);
            rst = pktctrl_read_message(
                endpointobj, &endpoint2expprog);
            if (rst < 0) {
                warn("Worker %" PRIuFAST32
                    " read message from endpoint failed,"
                    " skipping this endpoint", worker_id);
                goto fail_msgbuf;
            } else if (rst == 0) {
                assert(endpoint2expprog == NULL);
                debug("Worker %" PRIuFAST32
                    " endpoint message temporarily unavailable",
                    worker_id);
                // retry later
            } else {
                if (endpoint2expprog == NULL) { // eof
                    info("Worker %" PRIuFAST32
                        " got endpoint EOF, ending this session",
                        worker_id);
                    break;
                }
            }
        }

        if (endpoint_writable) {
            if (expprog2endpoint == NULL) {
                rst = pktctrl_flush_write(endpointobj);
                if (rst < 0) {
                    warn("Worker %" PRIuFAST32
                        " flush message to endpoint failed,"
                        " skipping this endpoint", worker_id);
                    goto fail_msgbuf;
                }
            } else {
                rst = pktctrl_write_message(
                    endpointobj, expprog2endpoint);
                if (rst < 0) {
                    warn("Worker %" PRIuFAST32
                        " write message to endpoint failed,"
                        " skipping this endpoint", worker_id);
                    goto fail_msgbuf;
                } else if (rst == 0) {
                    debug("Worker %" PRIuFAST32
                        " write message to endpoint failed temporarily",
                        worker_id);
                    // retry later
                } else {
                    assert(rst == 1);
                    free(expprog2endpoint);
                    expprog2endpoint = NULL;
                }
            }
        }

        if (expprog_readable) {
            assert(expprog2endpoint == NULL);
            rst = pktctrl_read_message(
                expprogobj, &expprog2endpoint);
            if (rst < 0) {
                warn("Worker %" PRIuFAST32
                    " read message from exp program failed,"
                    " skipping this endpoint", worker_id);
                goto fail_msgbuf;
            } else if (rst == 0) {
                assert(expprog2endpoint == NULL);
                debug("Worker %" PRIuFAST32
                    " exp program message temporarily unavailable",
                    worker_id);
                // retry
            } else {
                if (expprog2endpoint == NULL) { // eof
                    info("Worker %" PRIuFAST32
                        " got exp program EOF, ending this session",
                        worker_id);
                    break;
                }
            }
        }

        if (expprog_writable) {
            if (endpoint2expprog == NULL) {
                rst = pktctrl_flush_write(expprogobj);
                if (rst < 0) {
                    warn("Worker %" PRIuFAST32
                        " flush message to exp program failed,"
                        " skipping this endpoint", worker_id);
                    goto fail_msgbuf;
                }
            } else {
                rst = pktctrl_write_message(
                    expprogobj, endpoint2expprog);
                if (rst < 0) {
                    warn("Worker %" PRIuFAST32
                        " write message to exp program failed,"
                        " skipping this endpoint", worker_id);
                    goto fail_msgbuf;
                } else if (rst == 0) {
                    debug("Worker %" PRIuFAST32
                        " write message to exp program failed temporarily",
                        worker_id);
                    // retry later
                } else {
                    assert(rst == 1);
                    free(endpoint2expprog);
                    endpoint2expprog = NULL;
                }
            }
        }

        if (FD_ISSET(manager_fd, &rset)) { // eof
            info("Worker %" PRIuFAST32
                " detect manager EOF, ending this session",
                worker_id);
            break;
        }
    }

fail_msgbuf:
    free(endpoint2expprog);
    free(expprog2endpoint);
fail_expprogobj:
    pktctrl_close(expprogobj);
    close(fds[0]); // fds[1] already closed
    if (pid != 0) // signal exp prog by closing sock, then wait
        waitpid(pid, NULL, 0); // don't get child exit status
fail:
    pktctrl_close(endpointobj);
    close(endpoint_fd);
    close(manager_fd);
    return NULL;
}

int send_worker_msg(int sock, uint8_t msg) {
    ssize_t rst;

    assert(sock >= 0);

    trace("%s(sock:%d,msg:%u)",
        __func__, sock, (unsigned) msg);

    rst = write(sock, &msg, sizeof(msg));
    if (rst < 0) {
        if (errno == EAGAIN || errno == EWOULDBLOCK)
            return 0; // signal should retry
        return -1;
    }

    debug("Wrote worker msg %u into %d",
        (unsigned) msg, sock);

    assert(rst == sizeof(msg));
    return 1; // success
}

int recv_worker_msg(int sock, uint8_t * msg) {
    ssize_t rst;

    assert(sock >= 0 && msg != NULL);

    trace("%s(sock:%d)", __func__, sock);

    rst = read(sock, msg, sizeof(*msg));
    if (rst < 0) {
        if (errno == EAGAIN || errno == EWOULDBLOCK)
            return 0; // signal should retry
        return -1; // return error
    } else if (rst == 0) {
        debug("sock %d EOF", sock);
        *msg = WORKER_EOF;
        return 1;
    }

    debug("Read worker msg %u into %d",
        (unsigned) *msg, sock);

    assert(rst == sizeof(*msg));
    return 1; // success
}