// channel.c
//

#include <assert.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>

#include <pktconf.h>
#include <pktlab.h>

#include "console.h"
#include "util.h"

#include "channel.h"

//
// INTERNAL FUNCTION DECLARATIONS
//

static int get_pktlab_channel_hex (
    const uint8_t * bytes, size_t byteslen,
    struct pktlab_channel ** channel);

static void append_channel (
    struct pktlab_channel ** head,
    struct pktlab_channel ** tail,
    struct pktlab_channel * channel);

static int append_privchs (
    struct pktlab_channel ** head,
    struct pktlab_channel ** tail,
    const struct private_channel * privchs);

//
// EXPORTED FUNCTION DEFINITIONS
//

int get_pktlab_channels (
    const struct pktconf_sep_str_ll * strs,
    const struct private_channel * privchs,
    struct pktlab_channel ** channels)
{
    const struct pktconf_sep_str_entry * entry;
    struct pktlab_channel * channel;
    struct pktlab_channel * tail;
    uint8_t * bytes;
    size_t byteslen;
    int ret = -128;

    assert(channels != NULL);

    trace("%s(...)", __func__);

    tail = *channels = NULL;

    if (strs == NULL) {
        if (privchs == NULL)
            return -1; // missing private channel ll

        if (append_privchs(
                channels, &tail, privchs) < 0) {
            ret = -2;
            goto fail_channels;
        }

        return 0;
    }

    for (entry = strs->head; entry != NULL; entry = entry->next) {
        if (entry->entrylen == 0) {
            if (privchs == NULL) {
                ret = -1; // missing channel key
                goto fail_channels;
            }

            if (append_privchs(
                    channels, &tail, privchs) < 0) {
                ret = -2;
                goto fail_channels;
            }

            continue;
        }

        if (entry->entrylen % 2 != 0) {
            ret = -3;
            goto fail_channels;
        }

        if (entry->entrylen > SIZE_MAX) {
            ret = -4;
            goto fail_channels;
        }

        if (hexstr2bytes(
                entry->entry, entry->entrylen,
                &bytes, &byteslen) < 0) {
            ret = -5;
            goto fail_channels;
        }

        if (get_pktlab_channel_hex(
                bytes, byteslen, &channel) < 0) {
            ret = -6;
            free(bytes);
            goto fail_channels;
        }

        free(bytes);
        append_channel(channels, &tail, channel);
    }

    return 0;

fail_channels:
    free_pktlab_channels(*channels);
    return ret;
}

void free_pktlab_channels (
    struct pktlab_channel * channels)
{
    struct pktlab_channel * node;

    while (channels != NULL) {
        node = channels->next;
        free(channels);
        channels = node;
    }

    return;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

int get_pktlab_channel_hex (
    const uint8_t * bytes, size_t byteslen,
    struct pktlab_channel ** channel)
{
    assert(
        bytes != NULL && byteslen != 0 &&
        channel != NULL);

    if (byteslen > SIZE_MAX-sizeof(**channel) ||
        byteslen > UINT_FAST32_MAX)
        return -1; // overflow

    *channel = safe_malloc(byteslen+sizeof(**channel));
    (*channel)->channel = (void *) (*channel+1);
    (*channel)->channellen = byteslen;
    (*channel)->next = NULL;

    memcpy((*channel)->channel, bytes, byteslen);
    return 0;
}

void append_channel (
    struct pktlab_channel ** head,
    struct pktlab_channel ** tail,
    struct pktlab_channel * channel)
{
    assert(
        head != NULL && tail != NULL &&
        (*head == NULL) == (*tail == NULL) &&
        channel != NULL);

    channel->next = NULL;

    if (*head == NULL) {
        *head = channel;
        *tail = channel;
        return;
    }

    (*tail)->next = channel;
    *tail = channel;
    return;
}

int append_privchs (
    struct pktlab_channel ** head,
    struct pktlab_channel ** tail,
    const struct private_channel * privchs)
{
    const struct private_channel * privch;
    struct pktlab_channel * channel;

    assert(
        head != NULL && tail != NULL &&
        (*head == NULL) == (*tail == NULL) &&
        privchs != NULL);

    for (privch = privchs; privch != NULL; privch = privch->next) {
        if (get_pktlab_channel_hex(
                privch->channelid,
                sizeof(privch->channelid),
                &channel) < 0)
            return -1; // does not clean up past appended channels

        append_channel(head, tail, channel);
    }

    return 0;
}